"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

from __future__ import division
import time
import torch
import numpy as np
import cv2
import os
import sys
import random
import pickle as pkl
import argparse

from fmpose3d.lib.yolov3.util import *
from fmpose3d.lib.yolov3.darknet import Darknet
from fmpose3d.lib.yolov3 import preprocess
from fmpose3d.lib.checkpoint.download_checkpoints import get_checkpoint_path

cur_dir = os.path.dirname(os.path.realpath(__file__))
project_root = os.path.join(cur_dir, '../../../')
data_root = os.path.join(project_root, 'data/')


sys.path.insert(0, project_root)
sys.path.pop(0)


def prep_image(img, inp_dim):
    """
    Prepare image for inputting to the neural network.

    Returns a Variable
    """
    ori_img = img
    dim = ori_img.shape[1], ori_img.shape[0]
    img = cv2.resize(ori_img, (inp_dim, inp_dim))
    img_ = img[:, :, ::-1].transpose((2, 0, 1)).copy()
    img_ = torch.from_numpy(img_).float().div(255.0).unsqueeze(0)
    return img_, ori_img, dim


def write(x, img, colors):
    x = [int(i) for i in x]
    c1 = tuple(x[0:2])
    c2 = tuple(x[2:4])

    label = 'People {}'.format(0)
    color = (0, 0, 255)
    cv2.rectangle(img, c1, c2, color, 2)
    t_size = cv2.getTextSize(label, cv2.FONT_HERSHEY_PLAIN, 1, 1)[0]
    c2 = c1[0] + t_size[0] + 3, c1[1] + t_size[1] + 4
    cv2.rectangle(img, c1, c2, color, -1)
    cv2.putText(img, label, (c1[0], c1[1] + t_size[1] + 4), cv2.FONT_HERSHEY_PLAIN, 1, [225, 255, 255], 1)
    return img


def arg_parse():
    """"
    Parse arguements to the detect module

    """
    parser = argparse.ArgumentParser(description='YOLO v3 Cam Demo')
    parser.add_argument('--confidence', dest='confidence', type=float, default=0.70,
                        help='Object Confidence to filter predictions')
    parser.add_argument('--nms-thresh', dest='nms_thresh', type=float, default=0.4, help='NMS Threshold')
    parser.add_argument('--reso', dest='reso', default=416, type=int, help='Input resolution of the network. '
                        'Increase to increase accuracy. Decrease to increase speed. (160, 416)')
    parser.add_argument('-wf', '--weight-file', type=str, default=get_checkpoint_path('yolov3.weights'), 
                        help='The path of model weight file')
    parser.add_argument('-cf', '--cfg-file', type=str, default=cur_dir + '/cfg/yolov3.cfg', help='weight file')
    parser.add_argument('-a', '--animation', action='store_true', help='output animation')
    parser.add_argument('-i', '--image', type=str, default='lib/yolov3/data/dog-cycle-car.png',
                        help='The input video path')
    parser.add_argument('-np', '--num-person', type=int, default=1, help='number of estimated human poses. [1, 2]')
    parser.add_argument('--gpu', type=str, default='0', help='input video')
    
    parser.add_argument('--type', type=str, default='image', help='input type, only support image or video')
    parser.add_argument('--path', type=str, default='images/running.png', help='the path of your file')

    args, unknown = parser.parse_known_args()
    return args


def load_model(args=None, CUDA=None, inp_dim=416):
    if args is None:
        args = arg_parse()

    if CUDA is None:
        CUDA = torch.cuda.is_available()

    # Set up the neural network
    model = Darknet(args.cfg_file)
    model.load_weights(args.weight_file)
    # print("YOLOv3 network successfully loaded")

    model.net_info["height"] = inp_dim
    assert inp_dim % 32 == 0
    assert inp_dim > 32

    # If there's a GPU availible, put the model on GPU
    if CUDA:
        model.cuda()

    # Set the model in evaluation mode
    model.eval()

    return model


def yolo_human_det(img, model=None, reso=416, confidence=0.70):
    args = arg_parse()
    # args.reso = reso
    inp_dim = reso
    num_classes = 80

    CUDA = torch.cuda.is_available()
    if model is None:
        model = load_model(args, CUDA, inp_dim)

    if type(img) == str:
        assert os.path.isfile(img), 'The image path does not exist'
        img = cv2.imread(img)

    img, ori_img, img_dim = preprocess.prep_image(img, inp_dim)
    img_dim = torch.FloatTensor(img_dim).repeat(1, 2)

    with torch.no_grad():
        if CUDA:
            img_dim = img_dim.cuda()
            img = img.cuda()
        output = model(img, CUDA)
        output = write_results(output, confidence, num_classes, nms=True, nms_conf=args.nms_thresh, det_hm=True)

        if len(output) == 0:
            return None, None

        img_dim = img_dim.repeat(output.size(0), 1)
        scaling_factor = torch.min(inp_dim / img_dim, 1)[0].view(-1, 1)

        output[:, [1, 3]] -= (inp_dim - scaling_factor * img_dim[:, 0].view(-1, 1)) / 2
        output[:, [2, 4]] -= (inp_dim - scaling_factor * img_dim[:, 1].view(-1, 1)) / 2
        output[:, 1:5] /= scaling_factor

        for i in range(output.shape[0]):
            output[i, [1, 3]] = torch.clamp(output[i, [1, 3]], 0.0, img_dim[i, 0])
            output[i, [2, 4]] = torch.clamp(output[i, [2, 4]], 0.0, img_dim[i, 1])

    bboxs = []
    scores = []
    for i in range(len(output)):
        item = output[i]
        bbox = item[1:5].cpu().numpy()
        # convert float32 to .2f data
        bbox = [round(i, 2) for i in list(bbox)]
        score = item[5].cpu().numpy()
        bboxs.append(bbox)
        scores.append(score)
    scores = np.expand_dims(np.array(scores), 1)
    bboxs = np.array(bboxs)

    return bboxs, scores



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

import argparse
import math


def str2bool(v):
    """Convert string to boolean for argparse"""
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


class opts:
    def __init__(self):
        self.parser = argparse.ArgumentParser()

    def init(self):
        self.parser.add_argument("--model", default="", type=str)
        self.parser.add_argument("--layers", default=3, type=int)
        self.parser.add_argument("--channel", default=512, type=int)
        self.parser.add_argument("--d_hid", default=1024, type=int)
        self.parser.add_argument("--dataset", type=str, default="rat7m")
        self.parser.add_argument("-k", "--keypoints", default="cpn_ft_h36m_dbb", type=str)
        self.parser.add_argument("--data_augmentation", type=bool, default=False)
        self.parser.add_argument("--reverse_augmentation", type=bool, default=False)
        self.parser.add_argument("--test_augmentation", type=str2bool, default=False)
        self.parser.add_argument(
            "--test_augmentation_flip_hypothesis", type=str2bool, default=False
        )
        self.parser.add_argument("--test_augmentation_FlowAug", type=str2bool, default=False)
        self.parser.add_argument("--crop_uv", type=int, default=0)
        self.parser.add_argument("--root_path", type=str, default="Rat7M_data/")
        self.parser.add_argument("-a", "--actions", default="*", type=str)
        self.parser.add_argument("--downsample", default=1, type=int)
        self.parser.add_argument("--subset", default=1, type=float)
        self.parser.add_argument("-s", "--stride", default=1, type=int)
        self.parser.add_argument("--gpu", default="0", type=str, help="")
        self.parser.add_argument("--train", action="store_true")
        # self.parser.add_argument('--test', action='store_true')
        self.parser.add_argument("--test", type=int, default=1)  #
        self.parser.add_argument("--nepoch", type=int, default=41)  #
        self.parser.add_argument(
            "--batch_size", type=int, default=128, help="can be changed depending on your machine"
        )  # default 128
        self.parser.add_argument("--lr", type=float, default=1e-3)
        self.parser.add_argument("--lr_decay_large", type=float, default=0.5)
        self.parser.add_argument("--large_decay_epoch", type=int, default=20)
        self.parser.add_argument("--workers", type=int, default=8)
        self.parser.add_argument("-lrd", "--lr_decay", default=0.95, type=float)
        self.parser.add_argument("--frames", type=int, default=1)  #
        self.parser.add_argument(
            "--pad", type=int, default=175
        )  # 175  pad = (self.opt.frames-1) // 2
        self.parser.add_argument("--reload", action="store_true")
        self.parser.add_argument("--model_dir", type=str, default="")
        # Optional: load model class from a specific file path
        self.parser.add_argument("--model_path", type=str, default="")

        self.parser.add_argument("--post_refine_reload", action="store_true")
        self.parser.add_argument("--checkpoint", type=str, default="")
        self.parser.add_argument(
            "--previous_dir", type=str, default="./pre_trained_model/pretrained"
        )
        self.parser.add_argument("--saved_model_path", type=str, default="")

        self.parser.add_argument("--n_joints", type=int, default=26)
        self.parser.add_argument("--out_joints", type=int, default=26)
        self.parser.add_argument("--out_all", type=int, default=1)
        self.parser.add_argument("--in_channels", type=int, default=2)
        self.parser.add_argument("--out_channels", type=int, default=3)
        self.parser.add_argument("-previous_best_threshold", type=float, default=math.inf)
        self.parser.add_argument("-previous_name", type=str, default="")
        self.parser.add_argument(
            "--post_refine", action="store_true", help="if use post_refine model"
        )
        self.parser.add_argument(
            "-previous_post_refine_name", type=str, default="", help="save last saved model name"
        )
        self.parser.add_argument(
            "-norm", "--norm", default=0.01, type=float, metavar="N", help="constraint  of sparsity"
        )
        self.parser.add_argument("--train_views", type=int, nargs="+", default=[0, 1, 2, 3, 4, 5])
        self.parser.add_argument("--test_views", type=int, nargs="+", default=[0, 1, 2, 3, 4, 5])
        self.parser.add_argument("--token_dim", type=int, default=256)
        self.parser.add_argument("--create_time", type=str, default="")
        self.parser.add_argument("--filename", type=str, default="")
        self.parser.add_argument("--single", action="store_true")
        self.parser.add_argument("--reload_3d", action="store_true")

        #
        self.parser.add_argument("--create_file", type=int, default=1)
        self.parser.add_argument("--debug", action="store_true")
        self.parser.add_argument("--folder_name", type=str, default="")

        # param for refine
        self.parser.add_argument("--lr_refine", type=float, default=1e-5)
        self.parser.add_argument("--refine", action="store_true")
        self.parser.add_argument("--reload_refine", action="store_true")
        self.parser.add_argument("-previous_refine_name", type=str, default="")

        self.parser.add_argument("--sample_steps", type=int, default=3)
        # evaluation: run multiple sample steps at test time
        self.parser.add_argument("--eval_sample_steps", type=str, default="1,3,5,7,9")
        # allow multiple hypothesis counts, e.g. --num_hypothesis_list 1 3 5 7 9
        self.parser.add_argument("--num_hypothesis_list", type=str, default="1")
        self.parser.add_argument("--hypothesis_num", type=int, default=1)
        # number of best checkpoints to keep
        self.parser.add_argument("--num_saved_models", type=int, default=3)
        self.parser.add_argument("--sh_file", type=str, default="")
        # uncertainty-aware aggregation threshold factor

        self.parser.add_argument("--ua_k", type=float, default=0.9)
        self.parser.add_argument("--topk", type=int, default=3)
        self.parser.add_argument("--weight_softmax_tau", type=float, default=1.0)
        self.parser.add_argument("--exp_temp", type=float, default=0.002)
        self.parser.add_argument("--mode", type=str, default="exp")

        # mask joints
        self.parser.add_argument("--mask_prob", type=float, default=0.5)
        self.parser.add_argument("--masked_joints", type=str, default="12,13")

        # General arguments for animal3d
        self.parser.add_argument(
            "--train_dataset_path",
            type=str,
            nargs="+",
            default=["./dataset/animal3d/train.json"],
            help="List of training dataset paths (can specify multiple)",
        )
        self.parser.add_argument(
            "--test_dataset_path",
            type=str,
            nargs="+",
            default=["./dataset/animal3d/test.json"],
            help="List of test dataset paths (can specify multiple)",
        )

        # demo
        self.parser.add_argument(
            "--type", type=str, default="image", help="input type, only support image or video"
        )
        self.parser.add_argument(
            "--path", type=str, default="demo/images/running.png", help="the path of your file"
        )

        self.parser.set_defaults(align_r=True)
        self.parser.set_defaults(align_t=True)
        self.parser.set_defaults(align_s=True)
        self.parser.set_defaults(align_trj=True)
        self.parser.set_defaults(trj_align_r=True)
        self.parser.set_defaults(trj_align_s=True)
        self.parser.set_defaults(test_flip=True)
        self.parser.set_defaults(test_rot=True)

    def parse(self):
        self.init()

        self.opt = self.parser.parse_args()
        self.opt.pad = (self.opt.frames - 1) // 2

        self.opt.subjects_train = "S1,S5,S6,S7,S8"
        self.opt.subjects_test = "S9,S11"

        if self.opt.dataset == "h36m":
            self.opt.subjects_train = "S1,S5,S6,S7,S8"
            self.opt.subjects_test = "S9,S11"

            if self.opt.keypoints.startswith("sh") or self.opt.keypoints.startswith("hr"):
                self.opt.n_joints = 16
                self.opt.out_joints = 16

                self.opt.joints_left = [4, 5, 6, 10, 11, 12]
                self.opt.joints_right = [1, 2, 3, 13, 14, 15]
            else:
                self.opt.n_joints = 17
                self.opt.out_joints = 17

                self.opt.joints_left = [4, 5, 6, 11, 12, 13]
                self.opt.joints_right = [1, 2, 3, 14, 15, 16]

        elif self.opt.dataset == "rat7m":
            # Rat7M dataset configuration
            self.opt.n_joints = 20
            self.opt.out_joints = 20
            self.opt.joints_left = [8, 10, 11, 17, 18]  # HipL, ElbowL, ArmL, KneeL, ShinL
            self.opt.joints_right = [9, 14, 15, 16, 19]  # HipR, ElbowR, ArmR, KneeR, ShinR
            self.opt.root_joint = 4
        elif self.opt.dataset == "animal3d":
            # Animal3D dataset configuration

            """
            [0: 'left_eye', 1: 'right_eye', 2:'mouth_mid', 3:'left_front_paw',
            4: 'right_front_paw', 5:'left_back_paw', 6: 'right_back_paw', 7: 'tail_base',
            8: 'left_front_thigh', 9: 'right_front_thigh', 10: 'left_back_thigh', 11: 'right_back_thigh',
            12: 'left_shoulder', 13: 'right_shoulder', 14: 'left_front_knee', 15: 'right_front_knee',
            16: 'left_back_knee', 17: 'right_back_knee',18: 'neck', 19: 'tail_end',
            20: 'left_ear', 21: 'right_ear', 22: 'left_mouth', 23: 'right_mouth',
            24: 'nose', 25: 'tail_mid']

            skeleton pairs:
                pairs = [0, 24,  1, 24,  2, 24,  3, 14,  4, 15,  5, 16,  6, 17,  7, 18,  8, 12,  9, 13,  10, 7,  11, 7,
                12, 18,  13, 18,  14, 8,  15, 9,  16, 10,  17, 11,  18, 24,  19, 25,  20, 0,  21, 1,  22, 24,
                23, 24,  25, 7]

            """

            self.opt.n_joints = 26
            self.opt.out_joints = 26
            # Root joint: Body_Center (index 12) is a stable center point
            # Alternative: Hip_Center (index 13)
            self.opt.root_joint = 7  # Body_Center - most stable central point -- use tail_base here
            self.opt.joints_left = [0, 3, 5, 8, 10, 12, 14, 16, 20, 22]
            self.opt.joints_right = [1, 4, 6, 9, 11, 13, 15, 17, 21, 23]

        return self.opt



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

import argparse
import math


def str2bool(v):
    """Convert string to boolean for argparse"""
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


class opts:
    def __init__(self):
        self.parser = argparse.ArgumentParser()

    def init(self):
        self.parser.add_argument("--model", default="", type=str)
        self.parser.add_argument("--layers", default=3, type=int)
        self.parser.add_argument("--channel", default=512, type=int)
        self.parser.add_argument("--d_hid", default=1024, type=int)
        self.parser.add_argument("--dataset", type=str, default="h36m")
        self.parser.add_argument(
            "-k", "--keypoints", default="cpn_ft_h36m_dbb", type=str
        )
        self.parser.add_argument("--data_augmentation", type=bool, default=True)
        self.parser.add_argument("--reverse_augmentation", type=bool, default=False)
        self.parser.add_argument("--test_augmentation", type=str2bool, default=True)
        self.parser.add_argument(
            "--test_augmentation_flip_hypothesis", type=str2bool, default=False
        )
        self.parser.add_argument(
            "--test_augmentation_FlowAug", type=str2bool, default=False
        )
        self.parser.add_argument("--crop_uv", type=int, default=0)
        self.parser.add_argument("--root_path", type=str, default="dataset/")
        self.parser.add_argument("-a", "--actions", default="*", type=str)
        self.parser.add_argument("--downsample", default=1, type=int)
        self.parser.add_argument("--subset", default=1, type=float)
        self.parser.add_argument("-s", "--stride", default=1, type=int)
        self.parser.add_argument("--gpu", default="0", type=str, help="")
        self.parser.add_argument("--train", action="store_true")
        # self.parser.add_argument('--test', action='store_true')
        self.parser.add_argument("--test", type=int, default=1)  #
        self.parser.add_argument("--nepoch", type=int, default=41)  #
        self.parser.add_argument(
            "--batch_size",
            type=int,
            default=128,
            help="can be changed depending on your machine",
        )  # default 128
        self.parser.add_argument("--lr", type=float, default=1e-3)
        self.parser.add_argument("--lr_decay_large", type=float, default=0.5)
        self.parser.add_argument("--large_decay_epoch", type=int, default=5)
        self.parser.add_argument("--workers", type=int, default=8)
        self.parser.add_argument("-lrd", "--lr_decay", default=0.95, type=float)
        self.parser.add_argument("--frames", type=int, default=1)  #
        self.parser.add_argument(
            "--pad", type=int, default=175
        )  #  pad = (self.opt.frames-1) // 2
        self.parser.add_argument("--reload", action="store_true")
        self.parser.add_argument("--model_dir", type=str, default="")
        # Optional: load model class from a specific file path
        self.parser.add_argument("--model_path", type=str, default="")

        self.parser.add_argument("--post_refine_reload", action="store_true")
        self.parser.add_argument("--checkpoint", type=str, default="")
        self.parser.add_argument(
            "--previous_dir", type=str, default="./pre_trained_model/pretrained"
        )
        self.parser.add_argument("--saved_model_path", type=str, default="")

        self.parser.add_argument("--n_joints", type=int, default=17)
        self.parser.add_argument("--out_joints", type=int, default=17)
        self.parser.add_argument("--out_all", type=int, default=1)
        self.parser.add_argument("--in_channels", type=int, default=2)
        self.parser.add_argument("--out_channels", type=int, default=3)
        self.parser.add_argument(
            "-previous_best_threshold", type=float, default=math.inf
        )
        self.parser.add_argument("-previous_name", type=str, default="")
        self.parser.add_argument(
            "--post_refine", action="store_true", help="if use post_refine model"
        )
        self.parser.add_argument(
            "-previous_post_refine_name",
            type=str,
            default="",
            help="save last saved model name",
        )
        self.parser.add_argument(
            "-norm",
            "--norm",
            default=0.01,
            type=float,
            metavar="N",
            help="constraint  of sparsity",
        )
        self.parser.add_argument(
            "--train_views", type=int, nargs="+", default=[0, 1, 2, 3]
        )
        self.parser.add_argument(
            "--test_views", type=int, nargs="+", default=[0, 1, 2, 3]
        )

        self.parser.add_argument("--token_dim", type=int, default=256)  #
        self.parser.add_argument("--create_time", type=str, default="")
        self.parser.add_argument("--filename", type=str, default="")
        self.parser.add_argument("--single", action="store_true")
        self.parser.add_argument("--reload_3d", action="store_true")

        #
        self.parser.add_argument("--create_file", type=int, default=1)
        self.parser.add_argument("--debug", action="store_true")
        self.parser.add_argument("--folder_name", type=str, default="")

        # param for refine
        self.parser.add_argument("--lr_refine", type=float, default=1e-5)
        self.parser.add_argument("--refine", action="store_true")
        self.parser.add_argument("--reload_refine", action="store_true")
        self.parser.add_argument("-previous_refine_name", type=str, default="")

        self.parser.add_argument("--sample_steps", type=int, default=3)
        # evaluation: run multiple sample steps at test time
        self.parser.add_argument("--eval_multi_steps", action="store_true")
        self.parser.add_argument("--eval_sample_steps", type=str, default="1,3,5,7,9")
        # allow multiple hypothesis counts, e.g. --num_hypothesis_list 1 3 5 7 9
        self.parser.add_argument("--num_hypothesis_list", type=str, default="1")
        self.parser.add_argument("--hypothesis_num", type=int, default=1)
        # Classifier-Free Guidance scale (1.0 = no guidance, >1.0 = stronger conditioning)
        self.parser.add_argument("--guidance_scale", type=float, default=1.0)
        # number of best checkpoints to keep
        self.parser.add_argument("--num_saved_models", type=int, default=3)
        self.parser.add_argument("--sh_file", type=str, default="")
        # uncertainty-aware aggregation threshold factor

        self.parser.add_argument("--topk", type=int, default=3)
        self.parser.add_argument("--weight_softmax_tau", type=float, default=1.0)
        self.parser.add_argument("--exp_temp", type=float, default=0.002)
        self.parser.add_argument("--mode", type=str, default="exp")

        self.parser.add_argument("--opt_steps", type=int, default=2)

        # demo
        self.parser.add_argument('--type', type=str, default='image', help='input type, only support image or video')
        self.parser.add_argument('--path', type=str, default='demo/images/running.png', help='the path of your file')
    
    def parse(self):
        self.init()

        self.opt = self.parser.parse_args()
        self.opt.pad = (self.opt.frames - 1) // 2

        self.opt.subjects_train = "S1,S5,S6,S7,S8"
        self.opt.subjects_test = "S9,S11"

        self.opt.root_joint = 0
        if self.opt.dataset == "h36m":
            self.opt.subjects_train = "S1,S5,S6,S7,S8"
            self.opt.subjects_test = "S9,S11"

            if self.opt.keypoints.startswith("sh") or self.opt.keypoints.startswith(
                "hr"
            ):
                self.opt.n_joints = 16
                self.opt.out_joints = 16

                self.opt.joints_left = [4, 5, 6, 10, 11, 12]
                self.opt.joints_right = [1, 2, 3, 13, 14, 15]
            else:
                self.opt.n_joints = 17
                self.opt.out_joints = 17

                self.opt.joints_left = [4, 5, 6, 11, 12, 13]  # left side
                self.opt.joints_right = [1, 2, 3, 14, 15, 16]

        return self.opt



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

import numpy as np

class Graph():
    """ The Graph to model the skeletons of human body/hand/rat/animal

    Args:
        strategy (string): must be one of the follow candidates
        - spatial: Clustered Configuration

        layout (string): must be one of the follow candidates
        - 'hm36_gt': Ground truth structure of Human3.6M, with 17 joints per frame
        - 'animal3d': Skeleton structure for Animal3D dataset, with 26 joints per frame
        - 'rat7m': Skeleton structure for Rat7M dataset, with 20 joints per frame

        max_hop (int): the maximal distance between two connected nodes
        dilation (int): controls the spacing between the kernel points

    """

    def __init__(self,
                 layout, 
                 strategy,
                 pad=0,
                 max_hop=1,
                 dilation=1):

        self.max_hop = max_hop # 1
        self.dilation = dilation # 1
        self.seqlen = pad  # 1
        self.get_edge(layout)
        self.hop_dis = get_hop_distance(self.num_node, self.edge, max_hop=max_hop) # [17,17], adjacent=1, self=0, others=inf

        # get distance of each node to center
        self.dist_center = self.get_distance_to_center(layout)  # dist_center: distance from each node to joint 7
        self.get_adjacency(strategy)

    def get_distance_to_center(self,layout): 
        """
        :return: get the distance of each node to center
        For hm36_gt: center is joint 7
        For animal3d: center is joint 18 (neck, root joint)
        For rat7m: center is joint 4 (SpineM, root joint)
        """
        dist_center = np.zeros(self.num_node)
        if layout == 'hm36_gt':
            for i in range(self.seqlen):
                index_start = i*self.num_node_each
                dist_center[index_start+0 : index_start+7] = [1, 2, 3, 4, 2, 3, 4]
                dist_center[index_start+7 : index_start+11] = [0, 1, 2, 3]
                dist_center[index_start+11 : index_start+17] = [2, 3, 4, 2, 3, 4]
        elif layout == 'animal3d':
            # Animal3D: 26 joints, center is joint 18 (neck)
            # Calculate distance from each joint to neck (joint 18) along skeleton
            for i in range(self.seqlen):
                index_start = i*self.num_node_each
                # Distance from each joint to neck (18):
                # Head: 24(1)→18(0), eyes/mouth from nose, ears from eyes
                # Front legs: 18→shoulder→thigh→knee→paw
                # Hind legs: 18→tail_base→thigh→knee→paw
                # Tail: 18→tail_base→tail_mid→tail_end
                dist_center[index_start+0 : index_start+26] = [
                    2, 2, 2, 4, 4,  # 0-4: left_eye, right_eye, mouth_mid, left_front_paw, right_front_paw
                    4, 4, 1, 2, 2,  # 5-9: left_back_paw, right_back_paw, tail_base, left_front_thigh, right_front_thigh
                    2, 2, 1, 1, 3,  # 10-14: left_back_thigh, right_back_thigh, left_shoulder, right_shoulder, left_front_knee
                    3, 3, 3, 0, 3,  # 15-19: right_front_knee, left_back_knee, right_back_knee, neck(center), tail_end
                    3, 3, 3, 3, 1,  # 20-24: left_ear, right_ear, left_mouth, right_mouth, nose
                    2               # 25: tail_mid
                ]

        return dist_center

    def __str__(self):
        return self.A

    def graph_link_between_frames(self,base):
        """
        calculate graph link between frames given base nodes and seq_ind
        :param base:
        :return:
        """
        return [((front) + i*self.num_node_each, (back)+ i*self.num_node_each) for i in range(self.seqlen) for (front, back) in base] # Connect joints across frames.


    def basic_layout(self,neighbour_base, sym_base):
        """
        for generating basic layout time link selflink etc.
        neighbour_base: neighbour link per frame
        sym_base: symmetrical link(for body) or cross-link(for hand) per frame

        :return: link each node with itself
        """
        self.num_node = self.num_node_each * self.seqlen
        time_link = [(i * self.num_node_each + j, (i + 1) * self.num_node_each + j) for i in range(self.seqlen - 1) # for single frame, this is null
                     for j in range(self.num_node_each)]
        self.time_link_forward = [(i * self.num_node_each + j, (i + 1) * self.num_node_each + j) for i in
                                  range(self.seqlen - 1) 
                                  for j in range(self.num_node_each)]
        self.time_link_back = [((i + 1) * self.num_node_each + j, (i) * self.num_node_each + j) for i in
                               range(self.seqlen - 1)
                               for j in range(self.num_node_each)]

        self_link = [(i, i) for i in range(self.num_node)]

        self.neighbour_link_all = self.graph_link_between_frames(neighbour_base)

        self.sym_link_all = self.graph_link_between_frames(sym_base)

        return self_link, time_link

    def get_edge(self, layout):
        """
        get edge link of the graph
        la,ra: left/right arm (for rat: left/right front leg)
        ll/rl: left/right leg (for rat: left/right hind leg)
        cb: center bone (spine)
        """
        if layout == 'hm36_gt':
            self.num_node_each = 17

            neighbour_base = [(0, 1), (2, 1), (3, 2), (4, 0), (5, 4), (6, 5), 
                              (7, 0), (8, 7), (9, 8), (10, 9), (11, 8),
                              (12, 11), (13, 12), (14, 8), (15, 14), (16, 15)
                              ]
                        
            sym_base = [(6, 3), (5, 2), (4, 1), (11, 14), (12, 15), (13, 16)]  

            self_link, time_link = self.basic_layout(neighbour_base, sym_base) # self_link: node itself; time_link: 

            self.la, self.ra =[11, 12, 13], [14, 15, 16] # left and right arm
            self.ll, self.rl = [4, 5, 6], [1, 2, 3] # left and right leg
            self.cb = [0, 7, 8, 9, 10] # center bone
            self.part = [self.la, self.ra, self.ll, self.rl, self.cb]

            self.edge = self_link + self.neighbour_link_all + self.sym_link_all + time_link # len=39

            # center node of body/hand
            self.center = 8 - 1
            
        elif layout == 'animal3d':
            # Animal3D: 26 joints
            # Joint indices from animaldata.md:
            # 0: left_eye, 1: right_eye, 2: mouth_mid, 3: left_front_paw, 4: right_front_paw
            # 5: left_back_paw, 6: right_back_paw, 7: tail_base, 8: left_front_thigh, 9: right_front_thigh
            # 10: left_back_thigh, 11: right_back_thigh, 12: left_shoulder, 13: right_shoulder
            # 14: left_front_knee, 15: right_front_knee, 16: left_back_knee, 17: right_back_knee
            # 18: neck, 19: tail_end, 20: left_ear, 21: right_ear, 22: left_mouth, 23: right_mouth
            # 24: nose, 25: tail_mid
            # Root joint: 18 (neck - most stable central point)
            
            self.num_node_each = 26
            
            # Neighbour connections based on skeleton from animaldata.md
            neighbour_base = [
                # Head connections
                (24, 0),   # nose → left_eye
                (24, 1),   # nose → right_eye
                (1, 21),   # right_eye → right_ear
                (0, 20),   # left_eye → left_ear (corrected)
                (24, 2),   # nose → mouth_mid
                (2, 22),   # mouth_mid → left_mouth
                (2, 23),   # mouth_mid → right_mouth
                (24, 18),  # nose → neck
                
                # Upper body (neck/shoulders/front legs)
                (18, 12),  # neck → left_shoulder
                (18, 13),  # neck → right_shoulder
                (12, 8),   # left_shoulder → left_front_thigh
                (13, 9),   # right_shoulder → right_front_thigh
                (8, 14),   # left_front_thigh → left_front_knee
                (9, 15),   # right_front_thigh → right_front_knee
                (14, 3),   # left_front_knee → left_front_paw
                (15, 4),   # right_front_knee → right_front_paw
                
                # Spine and hind legs
                (18, 7),   # neck → tail_base
                (7, 10),   # tail_base → left_back_thigh
                (7, 11),   # tail_base → right_back_thigh
                (10, 16),  # left_back_thigh → left_back_knee
                (11, 17),  # right_back_thigh → right_back_knee
                (16, 5),   # left_back_knee → left_back_paw
                (17, 6),   # right_back_knee → right_back_paw
                
                # Tail
                (7, 25),   # tail_base → tail_mid
                (25, 19),  # tail_mid → tail_end
            ]
            
            # Symmetry links between left and right sides
            sym_base = [
                (0, 1),    # left_eye <-> right_eye
                (20, 21),  # left_ear <-> right_ear
                (22, 23),  # left_mouth <-> right_mouth
                (12, 13),  # left_shoulder <-> right_shoulder
                (8, 9),    # left_front_thigh <-> right_front_thigh
                (14, 15),  # left_front_knee <-> right_front_knee
                (3, 4),    # left_front_paw <-> right_front_paw
                (10, 11),  # left_back_thigh <-> right_back_thigh
                (16, 17),  # left_back_knee <-> right_back_knee
                (5, 6),    # left_back_paw <-> right_back_paw
            ]
            
            self_link, time_link = self.basic_layout(neighbour_base, sym_base)
            
            # Body parts for animal skeleton
            self.left_front = [12, 8, 14, 3]      # Left front leg
            self.right_front = [13, 9, 15, 4]     # Right front leg
            self.left_hind = [10, 16, 5]          # Left hind leg
            self.right_hind = [11, 17, 6]         # Right hind leg
            self.spine = [18, 7]                  # Neck and tail_base
            self.tail = [7, 25, 19]               # Tail
            self.head = [24, 0, 1, 2, 20, 21, 22, 23]  # Head
            self.cb = self.spine + self.tail + [12, 13]  # Center body
            
            # For compatibility with original structure
            self.la = self.left_front
            self.ra = self.right_front
            self.ll = self.left_hind
            self.rl = self.right_hind
            self.part = [self.la, self.ra, self.ll, self.rl, self.cb]
            
            self.edge = self_link + self.neighbour_link_all + self.sym_link_all + time_link
            
            # Center node: joint 18 (neck, root joint)
            self.center = 18
            
        else:
            raise ValueError("Do Not Exist This Layout.")

    def get_adjacency(self, strategy):
        valid_hop = range(0, self.max_hop + 1, self.dilation) # [0, 1]
        adjacency = np.zeros((self.num_node, self.num_node)) 
        for hop in valid_hop:
            adjacency[self.hop_dis == hop] = 1
        normalize_adjacency = normalize_digraph(adjacency) 

        if strategy == 'spatial':
            A = []
            for hop in valid_hop:
                a_root = np.zeros((self.num_node, self.num_node))
                a_close = np.zeros((self.num_node, self.num_node))
                a_further = np.zeros((self.num_node, self.num_node))
                a_sym = np.zeros((self.num_node, self.num_node))
                a_forward = np.zeros((self.num_node, self.num_node))
                a_back = np.zeros((self.num_node, self.num_node))
                for i in range(self.num_node):
                    for j in range(self.num_node):
                        if self.hop_dis[j, i] == hop: # 0:diagonal; 1:adjacent point
                            if (j,i) in self.sym_link_all or (i,j) in self.sym_link_all: # symmetrical node
                                a_sym[j, i] = normalize_adjacency[j, i]
                            elif (j,i) in self.time_link_forward:
                                a_forward[j, i] = normalize_adjacency[j, i]
                            elif (j,i) in self.time_link_back:
                                a_back[j, i] = normalize_adjacency[j, i]
                            elif self.dist_center[j] == self.dist_center[i]: 
                                a_root[j, i] = normalize_adjacency[j, i]
                            elif self.dist_center[j] > self.dist_center[i]:
                                a_close[j, i] = normalize_adjacency[j, i]
                            else:
                                a_further[j, i] = normalize_adjacency[j, i] 

                if hop == 0:
                    A.append(a_root)
                else:
                    A.append(a_close)
                    A.append(a_further)
                    A.append(a_sym)
                    if self.seqlen > 1: 
                        A.append(a_forward)
                        A.append(a_back)

            A = np.stack(A)
            self.A = A

        else:
            raise ValueError("Do Not Exist This Strategy")
            
def get_hop_distance(num_node, edge, max_hop=1): # Build adjacency matrix; neighbors set to 1.
    A = np.zeros((num_node, num_node))
    for i, j in edge:
        A[j, i] = 1
        A[i, j] = 1
    # compute hop steps
    hop_dis = np.zeros((num_node, num_node)) + np.inf
    transfer_mat = [np.linalg.matrix_power(A, d) for d in range(max_hop + 1)]# GET [I, A]; matrix_power computes powers (0th -> identity, 1st -> A)
    arrive_mat = (np.stack(transfer_mat) > 0) # [2,17,17]
    for d in range(max_hop, -1, -1): # preserve A(i,j)=1 while A(i,i)=0; neighbors=1, diagonal=0
        hop_dis[arrive_mat[d]] = d
    return hop_dis

def normalize_digraph(A):
    Dl = np.sum(A, 0) 
    num_node = A.shape[0]
    Dn = np.zeros((num_node, num_node))
    for i in range(num_node):
        if Dl[i] > 0:
            Dn[i, i] = Dl[i]**(-1)
    AD = np.dot(A, Dn)
    return AD

def normalize_undigraph(A):
    Dl = np.sum(A, 0)
    num_node = A.shape[0]
    Dn = np.zeros((num_node, num_node)) # 17,17 
    for i in range(num_node):
        if Dl[i] > 0:
            Dn[i, i] = Dl[i]**(-0.5)
    DAD = np.dot(np.dot(Dn, A), Dn)
    return DAD

if __name__=="__main__":
    # Test Human3.6M skeleton
    print("Testing Human3.6M skeleton (17 joints):")
    graph_h36m = Graph('hm36_gt', 'spatial', 1)
    print(f"  Adjacency matrix shape: {graph_h36m.A.shape}")
    print(f"  Center joint: {graph_h36m.center}")
    print(f"  Number of nodes: {graph_h36m.num_node}")
    
    # Test Animal3D skeleton
    print("\nTesting Animal3D skeleton (26 joints):")
    graph_animal = Graph('animal3d', 'spatial', 1)
    print(f"  Adjacency matrix shape: {graph_animal.A.shape}")
    print(f"  Center joint: {graph_animal.center} (neck)")
    print(f"  Number of nodes: {graph_animal.num_node}")
    print(f"  Body parts:")
    print(f"    - Left front leg: {graph_animal.left_front}")
    print(f"    - Right front leg: {graph_animal.right_front}")
    print(f"    - Left hind leg: {graph_animal.left_hind}")
    print(f"    - Right hind leg: {graph_animal.right_hind}")
    print(f"    - Head: {graph_animal.head}")
    print(f"    - Tail: {graph_animal.tail}")
    print(f"  Distance to center (joint 18): {graph_animal.dist_center}")
    
    # Test Rat7M skeleton
    print("\nTesting Rat7M skeleton (20 joints):")
    graph_rat = Graph('rat7m', 'spatial', 1)
    print(f"  Adjacency matrix shape: {graph_rat.A.shape}")
    print(f"  Center joint: {graph_rat.center}")
    print(f"  Number of nodes: {graph_rat.num_node}")
    print(f"  Body parts:")
    print(f"    - Left front leg: {graph_rat.left_front}")
    print(f"    - Right front leg: {graph_rat.right_front}")
    print(f"    - Left hind leg: {graph_rat.left_hind}")
    print(f"    - Right hind leg: {graph_rat.right_hind}")
    print(f"    - Spine: {graph_rat.spine}")
    print(f"  Distance to center (joint 4): {graph_rat.dist_center}")


"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

__version__ = "0.0.7"
__author__ = "Ti Wang, Xiaohang Yu, Mackenzie Weygandt Mathis"
__license__ = "Apache 2.0"

# Import key components for easy access
from .aggregation_methods import (
    average_aggregation,
    aggregation_select_single_best_hypothesis_by_2D_error,
    aggregation_RPEA_joint_level,
)

# Import 2D pose detection utilities
from .lib.hrnet.gen_kpts import gen_video_kpts
from .lib.preprocess import h36m_coco_format, revise_kpts

# Make commonly used classes/functions available at package level
__all__ = [
    # Aggregation methods
    "average_aggregation",
    "aggregation_select_single_best_hypothesis_by_2D_error",
    "aggregation_RPEA_joint_level",
    # 2D pose detection
    "gen_video_kpts",
    "h36m_coco_format",
    "revise_kpts",
    # Version
    "__version__",
]




"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

"""
Shared utilities for animal datasets and models.
"""



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

import numpy as np


class ChunkedGenerator:
    def __init__(
        self,
        batch_size,
        cameras,
        poses_3d,
        poses_2d,
        chunk_length=1,
        pad=0,
        causal_shift=0,
        shuffle=False,
        random_seed=1234,
        augment=False,
        reverse_aug=False,
        kps_left=None,
        kps_right=None,
        joints_left=None,
        joints_right=None,
        endless=False,
        out_all=False,
    ):
        assert poses_3d is None or len(poses_3d) == len(poses_2d), (
            len(poses_3d),
            len(poses_2d),
        )
        assert cameras is None or len(cameras) == len(poses_2d)

        pairs = []
        self.saved_index = {}
        start_index = 0

        for key in poses_2d.keys():
            assert poses_3d is None or poses_3d[key].shape[0] == poses_3d[key].shape[0]
            n_chunks = (poses_2d[key].shape[0] + chunk_length - 1) // chunk_length
            offset = (n_chunks * chunk_length - poses_2d[key].shape[0]) // 2
            bounds = np.arange(n_chunks + 1) * chunk_length - offset
            augment_vector = np.full(len(bounds - 1), False, dtype=bool)
            reverse_augment_vector = np.full(len(bounds - 1), False, dtype=bool)
            keys = np.tile(np.array(key).reshape([1, 3]), (len(bounds - 1), 1))
            pairs += list(
                zip(
                    keys,
                    bounds[:-1],
                    bounds[1:],
                    augment_vector,
                    reverse_augment_vector,
                )
            )
            if reverse_aug:
                pairs += list(
                    zip(
                        keys,
                        bounds[:-1],
                        bounds[1:],
                        augment_vector,
                        ~reverse_augment_vector,
                    )
                )
            if augment:
                if reverse_aug:
                    pairs += list(
                        zip(
                            keys,
                            bounds[:-1],
                            bounds[1:],
                            ~augment_vector,
                            ~reverse_augment_vector,
                        )
                    )
                else:
                    pairs += list(
                        zip(
                            keys,
                            bounds[:-1],
                            bounds[1:],
                            ~augment_vector,
                            reverse_augment_vector,
                        )
                    )

            end_index = start_index + poses_3d[key].shape[0]
            self.saved_index[key] = [start_index, end_index]
            start_index = start_index + poses_3d[key].shape[0]

        if cameras is not None:
            self.batch_cam = np.empty((batch_size, cameras[key].shape[-1]))

        if poses_3d is not None:
            self.batch_3d = np.empty(
                (
                    batch_size,
                    chunk_length,
                    poses_3d[key].shape[-2],
                    poses_3d[key].shape[-1],
                )
            )  # [B,1,17,3]
        self.batch_2d = np.empty(
            (
                batch_size,
                chunk_length + 2 * pad,
                poses_2d[key].shape[-2],
                poses_2d[key].shape[-1],
            )
        )

        self.num_batches = (len(pairs) + batch_size - 1) // batch_size
        self.batch_size = batch_size
        self.random = np.random.RandomState(random_seed)
        self.pairs = pairs
        self.shuffle = shuffle
        self.pad = pad
        self.causal_shift = causal_shift
        self.endless = endless
        self.state = None

        self.cameras = cameras
        if cameras is not None:
            self.cameras = cameras
        self.poses_3d = poses_3d
        self.poses_2d = poses_2d

        self.augment = augment
        self.kps_left = kps_left
        self.kps_right = kps_right
        self.joints_left = joints_left
        self.joints_right = joints_right
        self.out_all = out_all

    def num_frames(self):
        return self.num_batches * self.batch_size

    def random_state(self):
        return self.random

    def set_random_state(self, random):
        self.random = random

    def augment_enabled(self):
        return self.augment

    def next_pairs(self):
        if self.state is None:
            if self.shuffle:
                pairs = self.random.permutation(self.pairs)
            else:
                pairs = self.pairs
            return 0, pairs
        else:
            return self.state

    def get_batch(self, seq_i, start_3d, end_3d, flip, reverse):
        subject, action, cam_index = seq_i
        seq_name = (subject, action, int(cam_index))
        start_2d = start_3d - self.pad - self.causal_shift
        end_2d = end_3d + self.pad - self.causal_shift

        seq_2d = self.poses_2d[seq_name].copy()
        low_2d = max(start_2d, 0)
        high_2d = min(end_2d, seq_2d.shape[0])
        pad_left_2d = low_2d - start_2d
        pad_right_2d = end_2d - high_2d
        if pad_left_2d != 0 or pad_right_2d != 0:
            self.batch_2d = np.pad(
                seq_2d[low_2d:high_2d],
                ((pad_left_2d, pad_right_2d), (0, 0), (0, 0)),
                "edge",
            )
        else:
            self.batch_2d = seq_2d[low_2d:high_2d]

        if flip:
            self.batch_2d[:, :, 0] *= -1
            self.batch_2d[:, self.kps_left + self.kps_right] = self.batch_2d[
                :, self.kps_right + self.kps_left
            ]
        if reverse:
            self.batch_2d = self.batch_2d[::-1].copy()

        if self.poses_3d is not None:
            seq_3d = self.poses_3d[seq_name].copy()
            if self.out_all:
                low_3d = low_2d
                high_3d = high_2d
                pad_left_3d = pad_left_2d
                pad_right_3d = pad_right_2d
            else:
                low_3d = max(start_3d, 0)
                high_3d = min(end_3d, seq_3d.shape[0])
                pad_left_3d = low_3d - start_3d
                pad_right_3d = end_3d - high_3d
            if pad_left_3d != 0 or pad_right_3d != 0:
                self.batch_3d = np.pad(
                    seq_3d[low_3d:high_3d],
                    ((pad_left_3d, pad_right_3d), (0, 0), (0, 0)),
                    "edge",
                )
            else:
                self.batch_3d = seq_3d[low_3d:high_3d]

            if flip:
                self.batch_3d[:, :, 0] *= -1
                self.batch_3d[:, self.joints_left + self.joints_right] = self.batch_3d[
                    :, self.joints_right + self.joints_left
                ]
            if reverse:
                self.batch_3d = self.batch_3d[::-1].copy()

        if self.cameras is not None:
            self.batch_cam = self.cameras[seq_name].copy()
            if flip:
                self.batch_cam[2] *= -1
                self.batch_cam[7] *= -1

        if self.poses_3d is None and self.cameras is None:
            return None, None, self.batch_2d.copy(), action, subject, int(cam_index)
        elif self.poses_3d is not None and self.cameras is None:
            return (
                np.zeros(9),
                self.batch_3d.copy(),
                self.batch_2d.copy(),
                action,
                subject,
                int(cam_index),
            )
        elif self.poses_3d is None:
            return (
                self.batch_cam,
                None,
                self.batch_2d.copy(),
                action,
                subject,
                int(cam_index),
            )
        else:
            return (
                self.batch_cam,
                self.batch_3d.copy(),
                self.batch_2d.copy(),
                action,
                subject,
                int(cam_index),
            )



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

from __future__ import division

import torch
import numpy as np
import cv2
import os.path as osp
from fmpose3d.lib.yolov3.bbox import bbox_iou
from fmpose3d.lib.checkpoint.download_checkpoints import get_checkpoint_dir


def get_path(cur_file):
    cur_dir = osp.dirname(osp.realpath(cur_file))
    project_root = osp.join(cur_dir, '../../../')
    chk_root = get_checkpoint_dir()  # Use cache directory
    data_root = osp.join(project_root, 'data/')

    return project_root, chk_root, data_root, cur_dir


def count_parameters(model):
    return sum(p.numel() for p in model.parameters())


def count_learnable_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def convert2cpu(matrix):
    if matrix.is_cuda:
        return torch.FloatTensor(matrix.size()).copy_(matrix)
    else:
        return matrix


def predict_transform(prediction, inp_dim, anchors, num_classes, CUDA = True):
    batch_size = prediction.size(0)
    stride = inp_dim // prediction.size(2)
    grid_size = inp_dim // stride
    bbox_attrs = 5 + num_classes
    num_anchors = len(anchors)

    anchors = [(a[0]/stride, a[1]/stride) for a in anchors]

    prediction = prediction.view(batch_size, bbox_attrs*num_anchors, grid_size*grid_size)
    prediction = prediction.transpose(1, 2).contiguous()
    prediction = prediction.view(batch_size, grid_size*grid_size*num_anchors, bbox_attrs)

    # Sigmoid the  centre_X, centre_Y. and object confidencce
    prediction[:, :, 0] = torch.sigmoid(prediction[:, :, 0])
    prediction[:, :, 1] = torch.sigmoid(prediction[:, :, 1])
    prediction[:, :, 4] = torch.sigmoid(prediction[:, :, 4])

    # Add the center offsets
    grid_len = np.arange(grid_size)
    a, b = np.meshgrid(grid_len, grid_len)

    x_offset = torch.FloatTensor(a).view(-1, 1)
    y_offset = torch.FloatTensor(b).view(-1, 1)

    if CUDA:
        x_offset = x_offset.cuda()
        y_offset = y_offset.cuda()

    x_y_offset = torch.cat((x_offset, y_offset), 1).repeat(1, num_anchors).view(-1, 2).unsqueeze(0)

    prediction[:, :, :2] += x_y_offset

    # log space transform height and the width
    anchors = torch.FloatTensor(anchors)

    if CUDA:
        anchors = anchors.cuda()

    anchors = anchors.repeat(grid_size*grid_size, 1).unsqueeze(0)
    prediction[:, :, 2:4] = torch.exp(prediction[:, :, 2:4])*anchors

    # Softmax the class scores
    prediction[:, :, 5: 5 + num_classes] = torch.sigmoid((prediction[:, :, 5: 5 + num_classes]))

    prediction[:, :, :4] *= stride

    return prediction


def load_classes(namesfile):
    fp = open(namesfile, "r")
    names = fp.read().split("\n")[:-1]
    return names


def get_im_dim(im):
    im = cv2.imread(im)
    w, h = im.shape[1], im.shape[0]
    return w, h


def unique(tensor):
    tensor_np = tensor.cpu().numpy()
    unique_np = np.unique(tensor_np)
    unique_tensor = torch.from_numpy(unique_np)

    tensor_res = tensor.new(unique_tensor.shape)
    tensor_res.copy_(unique_tensor)
    return tensor_res


# ADD SOFT NMS
def write_results(prediction, confidence, num_classes, nms=True, nms_conf=0.4, det_hm=False):
    """
        https://blog.paperspace.com/how-to-implement-a-yolo-v3-object-detector-from-scratch-in-pytorch-part-4/
        prediction: (B x 10647 x 85)
        B: the number of images in a batch,
        10647: the number of bounding boxes predicted per image. (52×52+26×26+13×13)×3=10647
        85: the number of bounding box attributes. (c_x, c_y, w, h, object confidence, and 80 class scores)

        output: Num_obj × [img_index, x_1, y_1, x_2, y_2, object confidence, class_score, label_index]
    """

    conf_mask = (prediction[:, :, 4] > confidence).float().unsqueeze(2)
    prediction = prediction*conf_mask

    box_a = prediction.new(prediction.shape)
    box_a[:, :, 0] = (prediction[:, :, 0] - prediction[:, :, 2]/2)
    box_a[:, :, 1] = (prediction[:, :, 1] - prediction[:, :, 3]/2)
    box_a[:, :, 2] = (prediction[:, :, 0] + prediction[:, :, 2]/2)
    box_a[:, :, 3] = (prediction[:, :, 1] + prediction[:, :, 3]/2)
    prediction[:, :, :4] = box_a[:, :, :4]

    batch_size = prediction.size(0)

    output = prediction.new(1, prediction.size(2) + 1)
    write = False

    for ind in range(batch_size):
        # select the image from the batch
        image_pred = prediction[ind]

        # Get the class having maximum score, and the index of that class
        # Get rid of num_classes softmax scores
        # Add the class index and the class score of class having maximum score
        max_conf, max_conf_index = torch.max(image_pred[:, 5:5 + num_classes], 1)
        max_conf = max_conf.float().unsqueeze(1)
        max_conf_index = max_conf_index.float().unsqueeze(1)
        seq = (image_pred[:, :5], max_conf, max_conf_index)
        image_pred = torch.cat(seq, 1)  # image_pred:(10647, 7) 7:[x1, y1, x2, y2, obj_score, max_conf, max_conf_index]

        # Get rid of the zero entries
        non_zero_ind = (torch.nonzero(image_pred[:, 4]))
        image_pred__ = image_pred[non_zero_ind.squeeze(), :].view(-1, 7)

        # filters out people id
        if det_hm:
            cls_mask = (image_pred__[:, -1] == 0).float()
            class_mask_ind = torch.nonzero(cls_mask).squeeze()
            image_pred_ = image_pred__[class_mask_ind].view(-1, 7)

            if torch.sum(cls_mask) == 0:
                return image_pred_
        else:
            image_pred_ = image_pred__

        # Get the various classes detected in the image
        try:
            # img_classes = unique(image_pred_[:, -1])
            img_classes = torch.unique(image_pred_[:, -1], sorted=True).float()
        except:
             continue

        # We will do NMS classwise
        #  import ipdb;ipdb.set_trace()
        for cls in img_classes:
            # get the detections with one particular class
            cls_mask = image_pred_*(image_pred_[:, -1] == cls).float().unsqueeze(1)
            class_mask_ind = torch.nonzero(cls_mask[:, -2]).squeeze()
            image_pred_class = image_pred_[class_mask_ind].view(-1, 7)

            # sort the detections such that the entry with the maximum objectness
            # confidence is at the top
            conf_sort_index = torch.sort(image_pred_class[:, 4], descending=True)[1]
            image_pred_class = image_pred_class[conf_sort_index]
            idx = image_pred_class.size(0)

            #  from soft_NMS import soft_nms
            #  boxes = image_pred_class[:,:4]
            #  scores = image_pred_class[:, 4]
            #  k, N = soft_nms(boxes, scores, method=2)
            #  image_pred_class = image_pred_class[k]

            # if nms has to be done
            if nms:
                # For each detection
                for i in range(idx):
                    # Get the IOUs of all boxes that come after the one we are looking at
                    # in the loop
                    try:
                        ious = bbox_iou(image_pred_class[i].unsqueeze(0), image_pred_class[i+1:])
                    except ValueError:
                        break

                    except IndexError:
                        break

                    # Zero out all the detections that have IoU > threshold
                    iou_mask = (ious < nms_conf).float().unsqueeze(1)
                    image_pred_class[i+1:] *= iou_mask

                    #  Remove the zero entries
                    non_zero_ind = torch.nonzero(image_pred_class[:, 4]).squeeze()
                    image_pred_class = image_pred_class[non_zero_ind].view(-1, 7)

            # Concatenate the batch_id of the image to the detection
            # this helps us identify which image does the detection correspond to
            # We use a linear structure to hold ALL the detections from the batch
            # the batch_dim is flattened
            # batch is identified by extra batch column

            batch_ind = image_pred_class.new(image_pred_class.size(0), 1).fill_(ind)
            seq = batch_ind, image_pred_class
            if not write:
                output = torch.cat(seq, 1)
                write = True
            else:
                out = torch.cat(seq, 1)
                output = torch.cat((output, out))

    return output



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

import glob
import hashlib
import json
import os
import shutil

import numpy as np
import torch
from torch.autograd import Variable

def deterministic_random(min_value, max_value, data):
    digest = hashlib.sha256(data.encode()).digest()
    raw_value = int.from_bytes(digest[:4], byteorder="little", signed=False)
    return int(raw_value / (2**32 - 1) * (max_value - min_value)) + min_value

def mpjpe_cal(predicted, target):
    assert predicted.shape == target.shape
    return torch.mean(torch.norm(predicted - target, dim=len(target.shape) - 1))


def test_calculation(predicted, target, action, error_sum, data_type, subject):
    error_sum = mpjpe_by_action_p1(predicted, target, action, error_sum)
    error_sum = mpjpe_by_action_p2(predicted, target, action, error_sum)
    return error_sum


def mpjpe_by_action_p1(predicted, target, action, action_error_sum):
    assert predicted.shape == target.shape
    num = predicted.size(0)
    dist = torch.mean(
        torch.norm(predicted - target, dim=len(target.shape) - 1),
        dim=len(target.shape) - 2,
    )
    if len(set(list(action))) == 1:
        end_index = action[0].find(" ")
        if end_index != -1:
            action_name = action[0][:end_index]
        else:
            action_name = action[0]

        action_error_sum[action_name]["p1"].update(torch.mean(dist).item() * num, num)
    else:
        for i in range(num):
            end_index = action[i].find(" ")
            if end_index != -1:
                action_name = action[i][:end_index]
            else:
                action_name = action[i]

            action_error_sum[action_name]["p1"].update(dist[i].item(), 1)

    return action_error_sum


def p_mpjpe(predicted, target):  # p2, Procrustes analysis MPJPE
    assert predicted.shape == target.shape

    muX = np.mean(target, axis=1, keepdims=True)  # B,1,3
    muY = np.mean(predicted, axis=1, keepdims=True)  # B,1,3

    X0 = target - muX
    Y0 = predicted - muY

    normX = np.sqrt(np.sum(X0**2, axis=(1, 2), keepdims=True))  # B,1,1
    normY = np.sqrt(np.sum(Y0**2, axis=(1, 2), keepdims=True))

    X0 /= normX
    Y0 /= normY

    H = np.matmul(X0.transpose(0, 2, 1), Y0)
    U, s, Vt = np.linalg.svd(H)
    V = Vt.transpose(0, 2, 1)
    R = np.matmul(V, U.transpose(0, 2, 1))

    sign_detR = np.sign(np.expand_dims(np.linalg.det(R), axis=1))
    V[:, :, -1] *= sign_detR
    s[:, -1] *= sign_detR.flatten()
    R = np.matmul(V, U.transpose(0, 2, 1))

    tr = np.expand_dims(np.sum(s, axis=1, keepdims=True), axis=2)

    a = tr * normX / normY
    t = muX - a * np.matmul(muY, R)

    predicted_aligned = a * np.matmul(predicted, R) + t

    return np.mean(
        np.linalg.norm(predicted_aligned - target, axis=len(target.shape) - 1),
        axis=len(target.shape) - 2,
    )


def mpjpe_by_action_p2(predicted, target, action, action_error_sum):
    assert predicted.shape == target.shape
    num = predicted.size(0)
    pred = (
        predicted.detach()
        .cpu()
        .numpy()
        .reshape(-1, predicted.shape[-2], predicted.shape[-1])
    )  # B,17,3
    gt = (
        target.detach().cpu().numpy().reshape(-1, target.shape[-2], target.shape[-1])
    )  # # B,17,3
    dist = p_mpjpe(pred, gt)

    if len(set(list(action))) == 1:
        end_index = action[0].find(" ")
        if end_index != -1:
            action_name = action[0][:end_index]
        else:
            action_name = action[0]
        action_error_sum[action_name]["p2"].update(np.mean(dist) * num, num)
    else:
        for i in range(num):
            end_index = action[i].find(" ")
            if end_index != -1:
                action_name = action[i][:end_index]
            else:
                action_name = action[i]
            action_error_sum[action_name]["p2"].update(np.mean(dist), 1)

    return action_error_sum


def define_actions(action):

    actions = [
        "Directions",
        "Discussion",
        "Eating",
        "Greeting",
        "Phoning",
        "Photo",
        "Posing",
        "Purchases",
        "Sitting",
        "SittingDown",
        "Smoking",
        "Waiting",
        "WalkDog",
        "Walking",
        "WalkTogether",
    ]

    if action == "All" or action == "all" or action == "*":
        return actions

    if not action in actions:
        raise (ValueError, "Unrecognized action: %s" % action)

    return [action]


def define_error_list(actions):
    error_sum = {}
    error_sum.update(
        {
            actions[i]: {"p1": AccumLoss(), "p2": AccumLoss()}
            for i in range(len(actions))
        }
    )
    return error_sum


class AccumLoss(object):
    def __init__(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val
        self.count += n
        self.avg = self.sum / self.count


def get_varialbe(split, target):
    num = len(target)
    var = []
    if split == "train":
        for i in range(num):
            temp = (
                Variable(target[i], requires_grad=False)
                .contiguous()
                .type(torch.cuda.FloatTensor)
            )
            var.append(temp)
    else:
        for i in range(num):
            temp = Variable(target[i]).contiguous().cuda().type(torch.cuda.FloatTensor)
            var.append(temp)

    return var


def print_error(data_type, action_error_sum, is_train):
    mean_error_p1, mean_error_p2 = print_error_action(action_error_sum, is_train)

    return mean_error_p1, mean_error_p2


def print_error_action(action_error_sum, is_train):
    mean_error_each = {"p1": 0.0, "p2": 0.0}
    mean_error_all = {"p1": AccumLoss(), "p2": AccumLoss()}

    if is_train == 0:
        print("{0:=^12} {1:=^10} {2:=^8}".format("Action", "p#1 mm", "p#2 mm"))

    for action, value in action_error_sum.items():
        if is_train == 0:
            print("{0:<12} ".format(action), end="")

        mean_error_each["p1"] = action_error_sum[action]["p1"].avg * 1000.0
        mean_error_all["p1"].update(mean_error_each["p1"], 1)

        mean_error_each["p2"] = action_error_sum[action]["p2"].avg * 1000.0
        mean_error_all["p2"].update(mean_error_each["p2"], 1)

        if is_train == 0:
            print(
                "{0:>6.2f} {1:>10.2f}".format(
                    mean_error_each["p1"], mean_error_each["p2"]
                )
            )

    if is_train == 0:
        print(
            "{0:<12} {1:>6.4f} {2:>10.4f}".format(
                "Average", mean_error_all["p1"].avg, mean_error_all["p2"].avg
            )
        )

    return mean_error_all["p1"].avg, mean_error_all["p2"].avg


def save_model(previous_name, save_dir, epoch, data_threshold, model, model_name):
    # remove the old model
    if os.path.exists(previous_name):
        os.remove(previous_name)

    torch.save(
        model.state_dict(),
        "%s/%s_%d_%d.pth" % (save_dir, model_name, epoch, data_threshold * 100),
    )
    previous_name = "%s/%s_%d_%d.pth" % (
        save_dir,
        model_name,
        epoch,
        data_threshold * 100,
    )
    return previous_name


def save_top_N_models(
    previous_name,
    save_dir,
    epoch,
    data_threshold,
    model,
    model_name,
    num_saved_models=3,
):
    """
    Save a checkpoint if it belongs to the top-N best (by lower data_threshold).
    Maintains an index file '<model_name>_top_models.json' in save_dir.

    Returns the path of the last saved checkpoint if a new one was saved,
    otherwise returns previous_name unchanged.
    """
    os.makedirs(save_dir, exist_ok=True)
    ckpt_path = os.path.join(
        save_dir, f"{model_name}_{epoch}_{int(data_threshold * 100)}.pth"
    )
    index_path = os.path.join(save_dir, f"{model_name}_top_models.json")

    # load current list
    top_list = []
    if os.path.exists(index_path):
        with open(index_path, "r") as f:
            top_list = json.load(f)

    # decide if we should save
    should_save = False
    if len(top_list) < int(num_saved_models):
        should_save = True
    else:
        worst_item = max(top_list, key=lambda x: x.get("p1", float("inf")))
        if data_threshold < float(worst_item.get("p1", float("inf"))):
            should_save = True

    if not should_save:
        return previous_name

    # save new checkpoint
    torch.save(model.state_dict(), ckpt_path)

    # append and trim to N
    top_list.append(
        {"p1": float(data_threshold), "path": ckpt_path, "epoch": int(epoch)}
    )
    # sort ascending by p1 and keep best N
    top_list.sort(key=lambda x: x.get("p1", float("inf")))
    while len(top_list) > int(num_saved_models):
        removed = top_list.pop()  # remove worst (last after sort ascending)
        if os.path.exists(removed.get("path", "")):
            os.remove(removed["path"])

    # write back index
    with open(index_path, "w") as f:
        json.dump(top_list, f, indent=2)

    # update best marker to point to current best (lowest p1): append _best to the original name
    if len(top_list) > 0:
        best_src = top_list[0].get("path")
        if best_src and os.path.exists(best_src):
            root_name, ext = os.path.splitext(best_src)
            best_path = f"{root_name}_best{ext}"
            try:
                # ensure only one best exists: remove all existing *_best for this model_name
                pattern = os.path.join(save_dir, f"{model_name}_*_best.pth")
                for old_best in glob.glob(pattern):
                    try:
                        os.remove(old_best)
                    except Exception:
                        pass
                shutil.copy2(best_src, best_path)
            except Exception:
                pass

    return ckpt_path


def back_to_ori_uv(cropped_uv, bb_box):
    """
    for cropped uv, back to origial uv to help do the uvd->xyz operation
    :return:
    """
    N, T, V, _ = cropped_uv.size()
    uv = (cropped_uv + 1) * (bb_box[:, 2:].view(N, 1, 1, 2) / 2.0) + bb_box[
        :, 0:2
    ].view(N, 1, 1, 2)
    return uv


def get_uvd2xyz(uvd, gt_3D, cam):
    """
    transfer uvd to xyz

    :param uvd: N*T*V*3 (uv and z channel)
    :param gt_3D: N*T*V*3 (NOTE: V=0 is absolute depth value of root joint)

    :return: root-relative xyz results
    """
    N, T, V, _ = uvd.size()

    dec_out_all = uvd.view(-1, T, V, 3).clone()  # N*T*V*3
    root = gt_3D[:, :, 0, :].unsqueeze(-2).repeat(1, 1, V, 1).clone()  # N*T*V*3
    enc_in_all = uvd[:, :, :, :2].view(-1, T, V, 2).clone()  # N*T*V*2

    cam_f_all = cam[..., :2].view(-1, 1, 1, 2).repeat(1, T, V, 1)  # N*T*V*2
    cam_c_all = cam[..., 2:4].view(-1, 1, 1, 2).repeat(1, T, V, 1)  # N*T*V*2

    # change to global
    z_global = dec_out_all[:, :, :, 2]  # N*T*V
    z_global[:, :, 0] = root[:, :, 0, 2]
    z_global[:, :, 1:] = dec_out_all[:, :, 1:, 2] + root[:, :, 1:, 2]  # N*T*V
    z_global = z_global.unsqueeze(-1)  # N*T*V*1

    uv = enc_in_all - cam_c_all  # N*T*V*2
    xy = uv * z_global.repeat(1, 1, 1, 2) / cam_f_all  # N*T*V*2
    xyz_global = torch.cat((xy, z_global), -1)  # N*T*V*3
    xyz_offset = xyz_global - xyz_global[:, :, 0, :].unsqueeze(-2).repeat(
        1, 1, V, 1
    )  # N*T*V*3

    return xyz_offset


def sym_penalty(dataset, keypoints, pred_out):
    """
    get penalty for the symmetry of human body
    :return:
    """
    loss_sym = 0
    if dataset == "h36m":
        if keypoints.startswith("sh"):
            left_bone = [(0, 4), (4, 5), (5, 6), (8, 10), (10, 11), (11, 12)]
            right_bone = [(0, 1), (1, 2), (2, 3), (8, 13), (13, 14), (14, 15)]
        else:
            left_bone = [(0, 4), (4, 5), (5, 6), (8, 11), (11, 12), (12, 13)]
            right_bone = [(0, 1), (1, 2), (2, 3), (8, 14), (14, 15), (15, 16)]
        for (i_left, j_left), (i_right, j_right) in zip(left_bone, right_bone):
            left_part = pred_out[:, :, i_left] - pred_out[:, :, j_left]
            right_part = pred_out[:, :, i_right] - pred_out[:, :, j_right]
            loss_sym += torch.mean(
                torch.norm(left_part, dim=-1) - torch.norm(right_part, dim=-1)
            )
    elif dataset.startswith("STB"):
        loss_sym = 0
    return loss_sym


def project_to_2d(X, camera_params):
    """
    Project 3D points to 2D using the Human3.6M camera projection function.
    This is a differentiable and batched reimplementation of the original MATLAB script.

    Arguments:
    X -- 3D points in *camera space* to transform (N, *, 3)
    camera_params -- intrinsic parameteres (N, 2+2+3+2=9)
    """
    assert X.shape[-1] == 3  #  B,J,3
    assert len(camera_params.shape) == 2  # camera_params:[B,1,9]
    assert camera_params.shape[-1] == 9
    assert X.shape[0] == camera_params.shape[0]

    while len(camera_params.shape) < len(X.shape):
        camera_params = camera_params.unsqueeze(1)

    f = camera_params[..., :2]
    c = camera_params[..., 2:4]
    k = camera_params[..., 4:7]  # B,1,3
    p = camera_params[..., 7:]  # B,1,2

    XX = torch.clamp(X[..., :2] / X[..., 2:], min=-1, max=1)  # B,J,2
    r2 = torch.sum(XX[..., :2] ** 2, dim=len(XX.shape) - 1, keepdim=True)  # B, J, 1

    radial = 1 + torch.sum(
        k * torch.cat((r2, r2**2, r2**3), dim=len(r2.shape) - 1),
        dim=len(r2.shape) - 1,  # B,J,1
        keepdim=True,
    )
    tan = torch.sum(p * XX, dim=len(XX.shape) - 1, keepdim=True)  # B,J,1

    XXX = XX * (radial + tan) + p * r2  # B,J,2

    return f * XXX + c


def input_augmentation(input_2D, model):
    joints_left = [4, 5, 6, 11, 12, 13]
    joints_right = [1, 2, 3, 14, 15, 16]

    input_2D_non_flip = input_2D[:, 0]
    input_2D_flip = input_2D[:, 1]

    output_3D_non_flip = model(input_2D_non_flip)
    output_3D_flip = model(input_2D_flip)

    output_3D_flip[:, :, :, 0] *= -1
    output_3D_flip[:, :, joints_left + joints_right, :] = output_3D_flip[
        :, :, joints_right + joints_left, :
    ]

    output_3D = (output_3D_non_flip + output_3D_flip) / 2

    input_2D = input_2D_non_flip

    return input_2D, output_3D



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

import numpy as np
import torch.utils.data as data

from fmpose3d.common.camera import normalize_screen_coordinates, world_to_camera
from fmpose3d.common.generator import ChunkedGenerator
from fmpose3d.common.utils import deterministic_random


class Fusion(data.Dataset):
    def __init__(self, opt, dataset, root_path, train=True):
        self.data_type = opt.dataset  # h36m
        self.train = train
        self.keypoints_name = opt.keypoints
        self.root_path = root_path

        self.train_list = opt.subjects_train.split(",")
        self.test_list = opt.subjects_test.split(",")
        self.action_filter = None if opt.actions == "*" else opt.actions.split(",")
        self.downsample = opt.downsample
        self.subset = opt.subset
        self.stride = opt.stride
        self.crop_uv = opt.crop_uv
        self.test_aug = opt.test_augmentation
        self.pad = opt.pad  # 0
        if self.train:
            self.keypoints = self.prepare_data(dataset, self.train_list)
            self.cameras_train, self.poses_train, self.poses_train_2d = self.fetch(
                dataset, self.train_list, subset=self.subset, views=opt.train_views
            )

            self.generator = ChunkedGenerator(
                opt.batch_size // opt.stride,
                self.cameras_train,
                self.poses_train,
                self.poses_train_2d,
                self.stride,
                pad=self.pad,
                augment=opt.data_augmentation,
                reverse_aug=opt.reverse_augmentation,
                kps_left=self.kps_left,
                kps_right=self.kps_right,
                joints_left=self.joints_left,
                joints_right=self.joints_right,
                out_all=opt.out_all,
            )
            print("INFO: Training on {} frames".format(self.generator.num_frames()))
        else:
            self.keypoints = self.prepare_data(dataset, self.test_list)
            self.cameras_test, self.poses_test, self.poses_test_2d = self.fetch(
                dataset, self.test_list, subset=self.subset, views=opt.test_views
            )

            self.generator = ChunkedGenerator(
                opt.batch_size // opt.stride,
                self.cameras_test,
                self.poses_test,
                self.poses_test_2d,
                pad=self.pad,
                augment=False,
                kps_left=self.kps_left,
                kps_right=self.kps_right,
                joints_left=self.joints_left,
                joints_right=self.joints_right,
            )
            self.key_index = self.generator.saved_index
            print("INFO: Testing on {} frames".format(self.generator.num_frames()))

    def prepare_data(
        self, dataset, folder_list
    ):  # dataset: {'S1':{'action1':{'positions': positions,'cameras': self._cameras[subject]}},}
        for subject in folder_list:
            for action in dataset[subject].keys():
                anim = dataset[subject][action]

                positions_3d = []
                for cam in anim["cameras"]:
                    pos_3d = world_to_camera(
                        anim["positions"], R=cam["orientation"], t=cam["translation"]
                    )
                    pos_3d[:, 1:] -= pos_3d[:, :1]
                    positions_3d.append(pos_3d)
                anim["positions_3d"] = positions_3d

        keypoints = np.load(
            self.root_path
            + "data_2d_"
            + self.data_type
            + "_"
            + self.keypoints_name
            + ".npz",
            allow_pickle=True,
        )  # files: positions_2d.npy, metadata.npy
        keypoints_symmetry = keypoints["metadata"].item()[
            "keypoints_symmetry"
        ]  # {'0':[4,5,6,11,12,13],'1':[1,2,3,14,15,16]}

        self.kps_left, self.kps_right = list(keypoints_symmetry[0]), list(
            keypoints_symmetry[1]
        )
        self.joints_left, self.joints_right = list(
            dataset.skeleton().joints_left()
        ), list(dataset.skeleton().joints_right())
        keypoints = keypoints[
            "positions_2d"
        ].item()  # dict   {'Si':{'actioni':[np.shape(frames,17,2)*4]}}

        for subject in folder_list:
            assert (
                subject in keypoints
            ), "Subject {} is missing from the 2D detections dataset".format(subject)
            for action in dataset[subject].keys():
                assert (
                    action in keypoints[subject]
                ), "Action {} of subject {} is missing from the 2D detections dataset".format(
                    action, subject
                )
                for cam_idx in range(len(keypoints[subject][action])):

                    mocap_length = dataset[subject][action]["positions_3d"][
                        cam_idx
                    ].shape[0]
                    assert (
                        keypoints[subject][action][cam_idx].shape[0] >= mocap_length
                    )  #

                    if keypoints[subject][action][cam_idx].shape[0] > mocap_length:
                        keypoints[subject][action][cam_idx] = keypoints[subject][
                            action
                        ][cam_idx][:mocap_length]

        for subject in keypoints.keys():
            for action in keypoints[subject]:
                for cam_idx, kps in enumerate(keypoints[subject][action]):
                    cam = dataset.cameras()[subject][cam_idx]
                    if self.crop_uv == 0:
                        kps[..., :2] = normalize_screen_coordinates(
                            kps[..., :2], w=cam["res_w"], h=cam["res_h"]
                        )
                    keypoints[subject][action][cam_idx] = kps

        return keypoints  # dict {'Si':{'actioni':[np.shape(frames,17,2)*4]}}

    def fetch(
        self, dataset, subjects, views=[0, 1, 2, 3], subset=1, parse_3d_poses=True
    ):
        out_poses_3d = {}
        out_poses_2d = {}
        out_camera_params = {}

        for subject in subjects:
            for action in self.keypoints[subject].keys():
                if self.action_filter is not None:
                    found = False
                    for a in self.action_filter:
                        if action.startswith(a):
                            found = True
                            break
                    if not found:
                        continue

                poses_2d = self.keypoints[subject][action]
                for i in range(len(poses_2d)):
                    if i in views:  # select views; default views: 0 1 2 3
                        out_poses_2d[(subject, action, i)] = poses_2d[i]

                if subject in dataset.cameras():
                    cams = dataset.cameras()[subject]
                    assert len(cams) == len(poses_2d), "Camera count mismatch"
                    for i, cam in enumerate(cams):
                        if i in views and "intrinsic" in cam:
                            out_camera_params[(subject, action, i)] = cam["intrinsic"]

                if parse_3d_poses and "positions_3d" in dataset[subject][action]:
                    poses_3d = dataset[subject][action]["positions_3d"]
                    assert len(poses_3d) == len(poses_2d), "Camera count mismatch"
                    for i in range(len(poses_3d)):
                        if i in views:
                            out_poses_3d[(subject, action, i)] = poses_3d[i]

        if len(out_camera_params) == 0:
            out_camera_params = None
        if len(out_poses_3d) == 0:
            out_poses_3d = None

        stride = self.downsample
        if subset < 1:
            for key in out_poses_2d.keys():
                n_frames = int(
                    round(len(out_poses_2d[key]) // stride * subset) * stride
                )
                start = deterministic_random(
                    0,
                    len(out_poses_2d[key]) - n_frames + 1,
                    str(len(out_poses_2d[key])),
                )
                out_poses_2d[key] = out_poses_2d[key][start : start + n_frames : stride]
                if out_poses_3d is not None:
                    out_poses_3d[key] = out_poses_3d[key][
                        start : start + n_frames : stride
                    ]
        elif stride > 1:
            for key in out_poses_2d.keys():
                out_poses_2d[key] = out_poses_2d[key][::stride]
                if out_poses_3d is not None:
                    out_poses_3d[key] = out_poses_3d[key][::stride]

        return out_camera_params, out_poses_3d, out_poses_2d

    def __len__(self):
        return len(self.generator.pairs)

    def __getitem__(self, index):
        seq_name, start_3d, end_3d, flip, reverse = self.generator.pairs[index]

        cam, gt_3D, input_2D, action, subject, cam_ind = self.generator.get_batch(
            seq_name, start_3d, end_3d, flip, reverse
        )

        if self.train == False and self.test_aug:
            _, _, input_2D_aug, _, _, _ = self.generator.get_batch(
                seq_name, start_3d, end_3d, flip=True, reverse=reverse
            )

            input_2D = np.concatenate(
                (
                    np.expand_dims(input_2D, axis=0),
                    np.expand_dims(input_2D_aug, axis=0),
                ),
                0,
            )
        bb_box = np.array([0, 0, 1, 1])
        input_2D_update = input_2D

        scale = 1.0
        return cam, gt_3D, input_2D_update, action, subject, scale, bb_box, cam_ind



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

"""
Auto-download checkpoint files from Google Drive if they don't exist.
Downloads to user cache directory for better compatibility.
"""
import os
import sys
import platform

# Required checkpoint files
REQUIRED_FILES = ['yolov3.weights', 'pose_hrnet_w48_384x288.pth']

# Google Drive folder URL
GDRIVE_FOLDER_URL = "https://drive.google.com/drive/folders/1_ENAMOsPM7FXmdYRbkwbFHgzQq_B_NQA"

def get_cache_dir():
    """Get the platform-specific cache directory for fmpose checkpoints."""
    system = platform.system()
    
    if system == "Windows":
        # Windows: %LOCALAPPDATA%\fmpose\checkpoints
        base = os.environ.get("LOCALAPPDATA", os.path.expanduser("~"))
        cache_dir = os.path.join(base, "fmpose", "checkpoints")
    elif system == "Darwin":
        # macOS: ~/Library/Caches/fmpose/checkpoints
        cache_dir = os.path.join(os.path.expanduser("~"), "Library", "Caches", "fmpose", "checkpoints")
    else:
        # Linux and others: ~/.cache/fmpose/checkpoints
        xdg_cache = os.environ.get("XDG_CACHE_HOME", os.path.join(os.path.expanduser("~"), ".cache"))
        cache_dir = os.path.join(xdg_cache, "fmpose", "checkpoints")
    
    return cache_dir

def get_checkpoint_dir():
    """Get the checkpoint directory path (creates if not exists)."""
    cache_dir = get_cache_dir()
    os.makedirs(cache_dir, exist_ok=True)
    return cache_dir

def get_checkpoint_path(filename):
    """Get the full path to a checkpoint file."""
    return os.path.join(get_checkpoint_dir(), filename)

def install_gdown():
    """Install gdown if not available."""
    try:
        import gdown
        return gdown
    except ImportError:
        print("Installing gdown for Google Drive downloads...")
        os.system(f"{sys.executable} -m pip install gdown -q")
        import gdown
        return gdown

def download_folder(output_dir):
    """Download all files from the Google Drive folder."""
    gdown = install_gdown()
    
    print(f"Downloading checkpoint files from Google Drive folder...")
    print(f"URL: {GDRIVE_FOLDER_URL}")
    print(f"Saving to: {output_dir}\n")
    
    try:
        gdown.download_folder(GDRIVE_FOLDER_URL, output=output_dir, quiet=False)
        return True
    except Exception as e:
        print(f"Folder download failed: {e}")
        return False

def ensure_checkpoints():
    """Check and download missing checkpoint files."""
    checkpoint_dir = get_checkpoint_dir()
    
    # Check which files are missing
    missing_files = []
    for filename in REQUIRED_FILES:
        filepath = os.path.join(checkpoint_dir, filename)
        if not os.path.exists(filepath):
            missing_files.append(filename)
    
    if not missing_files:
        return True
    
    print(f"\n{'='*60}")
    print("Missing checkpoint files detected:")
    for f in missing_files:
        print(f"  - {f}")
    print(f"{'='*60}")
    print("Starting auto-download from Google Drive...")
    print(f"Cache directory: {checkpoint_dir}")
    print(f"{'='*60}\n")
    
    # Try to download the folder
    success = download_folder(checkpoint_dir)
    
    # Verify all files are present
    still_missing = []
    for filename in REQUIRED_FILES:
        filepath = os.path.join(checkpoint_dir, filename)
        if not os.path.exists(filepath):
            still_missing.append(filename)
    
    if not still_missing:
        print(f"\n{'='*60}")
        print("All checkpoint files downloaded successfully!")
        print(f"{'='*60}\n")
        return True
    else:
        print(f"\n{'='*60}")
        print("WARNING: Some files are still missing:")
        for f in still_missing:
            print(f"  - {f}")
        print(f"\nPlease download manually from:")
        print(GDRIVE_FOLDER_URL)
        print(f"And place them in: {checkpoint_dir}")
        print(f"{'='*60}\n")
        return False

if __name__ == "__main__":
    ensure_checkpoints()
    print(f"\nCheckpoint directory: {get_checkpoint_dir()}")



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

import torch
from fmpose3d.common.utils import project_to_2d

def average_aggregation(list_hypothesis):
    return torch.mean(torch.stack(list_hypothesis), dim=0)


def aggregation_select_single_best_hypothesis_by_2D_error(args,
    list_hypothesis, batch_cam, input_2D, gt_3D
):
    """
    Select per-joint 3D from the hypothesis whose 2D projection yields minimal L2 error.

    Args:
        list_hypothesis: list of (B,1,J,3) tensors
        batch_cam: (B, 9) or (B, 1, 9) intrinsics [f(2), c(2), k(3), p(2)]
        input_2D: (B, F, J, 2) 2D joints in image coordinates
        gt_3D: (B, F, J, 3) used for shape metadata only
    Returns:
        (B,1,J,3) aggregated 3D pose with joint 0 set to 0
    """
    if len(list_hypothesis) == 0:
        raise ValueError("list_hypothesis is empty")

    device = list_hypothesis[0].device
    dtype = list_hypothesis[0].dtype

    # Shapes
    B = gt_3D.size(0)
    J = gt_3D.size(2)
    F = gt_3D.size(1)
    assert F >= 1, "Expected at least one frame"

    # Stack hypotheses: (H,B,1,J,3) -> (B,H,J,3)
    stack = torch.stack(list_hypothesis, dim=0)  # (H,B,1,J,3)
    X_hbj3 = stack[:, :, 0, :, :]  # (H,B,J,3)
    X_bhj3 = X_hbj3.transpose(0, 1).contiguous()  # (B,H,J,3)
    H = X_bhj3.size(1)

    # Prepare camera params: (B,9)
    if batch_cam.dim() == 3 and batch_cam.size(1) == 1:
        cam_b9 = batch_cam[:, 0, :].contiguous()
    elif batch_cam.dim() == 2 and batch_cam.size(1) == 9:
        cam_b9 = batch_cam
    else:
        cam_b9 = batch_cam.view(B, -1)
    assert cam_b9.size(1) == 9, f"camera params should be 9-dim, got {cam_b9.size()}"

    # Target 2D at the same frame index as 3D selection (args.pad)
    # input_2D: (B,F,J,2) -> (B,J,2)
    target_2d = input_2D[:, getattr(args, "pad", 0)].contiguous()  # (B,J,2)

    # Convert hypotheses from root-relative to absolute camera coordinates using GT root
    # Root at frame args.pad: (B,3)
    gt_root = gt_3D[:, getattr(args, "pad", 0), 0, :].contiguous()  # (B,3)
    X_abs = X_bhj3.clone()
    X_abs[:, :, 1:, :] = X_abs[:, :, 1:, :] + gt_root.unsqueeze(1).unsqueeze(1)
    X_abs[:, :, 0, :] = gt_root.unsqueeze(1)

    # Vectorized projection for all hypotheses in absolute coordinates
    # (B,H,J,3) -> (B*H,J,3)
    X_flat = X_abs.view(B * H, J, 3)
    cam_rep = cam_b9.repeat_interleave(H, dim=0)  # (B*H,9)

    # project_to_2d expects last dim=3 and cam (N,9)
    # Returns normalized coordinates (when crop_uv=0) because camera params are normalized
    proj2d_flat = project_to_2d(X_flat, cam_rep)  # (B*H,J,2) normalized coordinates
    proj2d_bhj = proj2d_flat.view(B, H, J, 2)

    # Per-hypothesis per-joint 2D error (both in normalized coordinates)
    diff = proj2d_bhj - target_2d.unsqueeze(1)  # (B,H,J,2)
    dist = torch.norm(diff, dim=-1)  # (B,H,J)

    # Exclude root joint (0) due to undefined depth when using root-relative 3D
    dist[:, :, 0] = float("inf")

    # Argmin across hypotheses per joint
    best_h = torch.argmin(dist, dim=1)  # (B,J)

    # Gather 3D using advanced indexing (return root-relative coordinates)
    b_idx = torch.arange(B, device=device).unsqueeze(1).expand(B, J)  # (B,J)
    j_idx = torch.arange(J, device=device).unsqueeze(0).expand(B, J)  # (B,J)
    selected_bj3 = X_bhj3[b_idx, best_h, j_idx, :]  # (B,J,3)

    agg = selected_bj3.unsqueeze(1).to(dtype=dtype)
    agg[:, :, 0, :] = 0
    return agg


def aggregation_RPEA_joint_level(
    args, list_hypothesis, batch_cam, input_2D, gt_3D, topk=3
):
    """
    Select per-joint 3D from the hypothesis whose 2D projection yields minimal L2 error.

    Args:
        list_hypothesis: list of (B,1,J,3) tensors
        batch_cam: (B, 9) or (B, 1, 9) intrinsics [f(2), c(2), k(3), p(2)]
        input_2D: (B, F, J, 2) 2D joints in image coordinates
        gt_3D: (B, F, J, 3) used for shape metadata only
    Returns:
        (B,1,J,3) aggregated 3D pose with joint 0 set to 0
    """
    if len(list_hypothesis) == 0:
        raise ValueError("list_hypothesis is empty")

    device = list_hypothesis[0].device
    dtype = list_hypothesis[0].dtype

    # Shapes
    B = gt_3D.size(0)
    J = gt_3D.size(2)
    F = gt_3D.size(1)
    assert F >= 1, "Expected at least one frame"

    # Stack hypotheses: (H,B,1,J,3) -> (B,H,J,3)
    stack = torch.stack(list_hypothesis, dim=0)  # (H,B,1,J,3)
    X_hbj3 = stack[:, :, 0, :, :]  # (H,B,J,3)
    X_bhj3 = X_hbj3.transpose(0, 1).contiguous()  # (B,H,J,3)
    H = X_bhj3.size(1)

    # Prepare camera params: (B,9)
    if batch_cam.dim() == 3 and batch_cam.size(1) == 1:
        cam_b9 = batch_cam[:, 0, :].contiguous()
    elif batch_cam.dim() == 2 and batch_cam.size(1) == 9:
        cam_b9 = batch_cam
    else:
        cam_b9 = batch_cam.view(B, -1)
    assert cam_b9.size(1) == 9, f"camera params should be 9-dim, got {cam_b9.size()}"

    # Target 2D at the same frame index as 3D selection (args.pad)
    # input_2D: (B,F,J,2) -> (B,J,2)
    target_2d = input_2D[:, getattr(args, "pad", 0)].contiguous()  # (B,J,2)

    # Convert hypotheses from root-relative to absolute camera coordinates using GT root
    # Root at frame args.pad: (B,3)
    gt_root = gt_3D[:, getattr(args, "pad", 0), 0, :].contiguous()  # (B,3)
    X_abs = X_bhj3.clone()
    X_abs[:, :, 1:, :] = X_abs[:, :, 1:, :] + gt_root.unsqueeze(1).unsqueeze(1)
    X_abs[:, :, 0, :] = gt_root.unsqueeze(1)

    # Vectorized projection for all hypotheses in absolute coordinates
    # (B,H,J,3) -> (B*H,J,3)
    X_flat = X_abs.view(B * H, J, 3)
    cam_rep = cam_b9.repeat_interleave(H, dim=0)  # (B*H,9)

    # project_to_2d expects last dim=3 and cam (N,9)
    proj2d_flat = project_to_2d(X_flat, cam_rep)  # (B*H,J,2)
    proj2d_bhj = proj2d_flat.view(B, H, J, 2)

    # Per-hypothesis per-joint 2D error
    diff = proj2d_bhj - target_2d.unsqueeze(1)  # (B,H,J,2)
    dist = torch.norm(diff, dim=-1)  # (B,H,J)

    # For root joint (0), avoid NaNs in softmax by setting equal distances
    # This yields uniform weights at the root (we set root to 0 later anyway)
    dist[:, :, 0] = 0.0

    # Convert 2D losses to weights using softmax over top-k hypotheses per joint
    tau = float(getattr(args, "weight_softmax_tau", 1.0))
    H = dist.size(1)
    k = int(getattr(args, "topk", None))
    # print("k:", k)
    # k = int(H//2)+1
    k = max(1, min(k, H))

    # top-k smallest distances along hypothesis dim
    topk_vals, topk_idx = torch.topk(dist, k=k, dim=1, largest=False)  # (B,k,J)

    # Weight calculation method ; weight_method = 'exp'
    temp = args.exp_temp
    max_safe_val = temp * 20
    topk_vals_clipped = torch.clamp(topk_vals, max=max_safe_val)
    exp_vals = torch.exp(-topk_vals_clipped / temp)
    exp_sum = exp_vals.sum(dim=1, keepdim=True)
    topk_weights = exp_vals / torch.clamp(exp_sum, min=1e-10)
    nan_mask = torch.isnan(topk_weights).any(dim=1, keepdim=True)
    uniform_weights = torch.ones_like(topk_weights) / k
    topk_weights = torch.where(
        nan_mask.expand_as(topk_weights), uniform_weights, topk_weights
    )

    # scatter back to full H with zeros elsewhere
    weights = torch.zeros_like(dist)  # (B,H,J)
    weights.scatter_(1, topk_idx, topk_weights)

    # Weighted sum of root-relative 3D hypotheses per joint
    weights_exp = weights.unsqueeze(-1)  # (B,H,J,1)
    weighted_bj3 = torch.sum(X_bhj3 * weights_exp, dim=1)  # (B,J,3)

    # Assemble output (B,1,J,3) and enforce root at origin
    agg = weighted_bj3.unsqueeze(1).to(dtype=dtype)
    agg[:, :, 0, :] = 0
    return agg


"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

import json
import numpy as np
import os

h36m_coco_order = [9, 11, 14, 12, 15, 13, 16, 4, 1, 5, 2, 6, 3]
coco_order = [0, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]
spple_keypoints = [10, 8, 0, 7]

def coco_h36m(keypoints):
    temporal = keypoints.shape[0]
    keypoints_h36m = np.zeros_like(keypoints, dtype=np.float32)
    htps_keypoints = np.zeros((temporal, 4, 2), dtype=np.float32)

    # htps_keypoints: head, thorax, pelvis, spine
    htps_keypoints[:, 0, 0] = np.mean(keypoints[:, 1:5, 0], axis=1, dtype=np.float32)
    htps_keypoints[:, 0, 1] = np.sum(keypoints[:, 1:3, 1], axis=1, dtype=np.float32) - keypoints[:, 0, 1]
    htps_keypoints[:, 1, :] = np.mean(keypoints[:, 5:7, :], axis=1, dtype=np.float32)
    htps_keypoints[:, 1, :] += (keypoints[:, 0, :] - htps_keypoints[:, 1, :]) / 3

    htps_keypoints[:, 2, :] = np.mean(keypoints[:, 11:13, :], axis=1, dtype=np.float32)
    htps_keypoints[:, 3, :] = np.mean(keypoints[:, [5, 6, 11, 12], :], axis=1, dtype=np.float32)

    keypoints_h36m[:, spple_keypoints, :] = htps_keypoints
    keypoints_h36m[:, h36m_coco_order, :] = keypoints[:, coco_order, :]

    keypoints_h36m[:, 9, :] -= (keypoints_h36m[:, 9, :] - np.mean(keypoints[:, 5:7, :], axis=1, dtype=np.float32)) / 4
    keypoints_h36m[:, 7, 0] += 2*(keypoints_h36m[:, 7, 0] - np.mean(keypoints_h36m[:, [0, 8], 0], axis=1, dtype=np.float32))
    keypoints_h36m[:, 8, 1] -= (np.mean(keypoints[:, 1:3, 1], axis=1, dtype=np.float32) - keypoints[:, 0, 1])*2/3

    # half body: the joint of ankle and knee equal to hip
    # keypoints_h36m[:, [2, 3]] = keypoints_h36m[:, [1, 1]]
    # keypoints_h36m[:, [5, 6]] = keypoints_h36m[:, [4, 4]]

    valid_frames = np.where(np.sum(keypoints_h36m.reshape(-1, 34), axis=1) != 0)[0]
    
    return keypoints_h36m, valid_frames


def h36m_coco_format(keypoints, scores):
    assert len(keypoints.shape) == 4 and len(scores.shape) == 3

    h36m_kpts = []
    h36m_scores = []
    valid_frames = []

    for i in range(keypoints.shape[0]):
        kpts = keypoints[i]
        score = scores[i]

        new_score = np.zeros_like(score, dtype=np.float32)

        if np.sum(kpts) != 0.:
            kpts, valid_frame = coco_h36m(kpts)
            h36m_kpts.append(kpts)
            valid_frames.append(valid_frame)

            new_score[:, h36m_coco_order] = score[:, coco_order]
            new_score[:, 0] = np.mean(score[:, [11, 12]], axis=1, dtype=np.float32)
            new_score[:, 8] = np.mean(score[:, [5, 6]], axis=1, dtype=np.float32)
            new_score[:, 7] = np.mean(new_score[:, [0, 8]], axis=1, dtype=np.float32)
            new_score[:, 10] = np.mean(score[:, [1, 2, 3, 4]], axis=1, dtype=np.float32)

            h36m_scores.append(new_score)

    h36m_kpts = np.asarray(h36m_kpts, dtype=np.float32) 
    h36m_scores = np.asarray(h36m_scores, dtype=np.float32)

    return h36m_kpts, h36m_scores, valid_frames


def revise_kpts(h36m_kpts, h36m_scores, valid_frames):

    new_h36m_kpts = np.zeros_like(h36m_kpts)
    for index, frames in enumerate(valid_frames):
        kpts = h36m_kpts[index, frames]
        score = h36m_scores[index, frames]

        index_frame = np.where(np.sum(score < 0.3, axis=1) > 0)[0]

        for frame in index_frame:
            less_threshold_joints = np.where(score[frame] < 0.3)[0]

            intersect = [i for i in [2, 3, 5, 6] if i in less_threshold_joints]

            if [2, 3, 5, 6] == intersect:
                kpts[frame, [2, 3, 5, 6]] = kpts[frame, [1, 1, 4, 4]]
            elif [2, 3, 6] == intersect:
                kpts[frame, [2, 3, 6]] = kpts[frame, [1, 1, 5]]
            elif [3, 5, 6] == intersect:
                kpts[frame, [3, 5, 6]] = kpts[frame, [2, 4, 4]]
            elif [3, 6] == intersect:
                kpts[frame, [3, 6]] = kpts[frame, [2, 5]]
            elif [3] == intersect:
                kpts[frame, 3] = kpts[frame, 2]
            elif [6] == intersect:
                kpts[frame, 6] = kpts[frame, 5]
            else:
                continue

        new_h36m_kpts[index, frames] = kpts

    return new_h36m_kpts





"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

from __future__ import division

import torch
import numpy as np
import cv2
from PIL import Image


def letterbox_image(img, inp_dim):
    '''resize image with unchanged aspect ratio using padding'''
    img_w, img_h = img.shape[1], img.shape[0]
    w, h = inp_dim
    new_w = int(img_w * min(w/img_w, h/img_h))
    new_h = int(img_h * min(w/img_w, h/img_h))
    resized_image = cv2.resize(img, (new_w, new_h), interpolation=cv2.INTER_CUBIC)

    canvas = np.full((inp_dim[1], inp_dim[0], 3), 128)

    canvas[(h - new_h) // 2:(h - new_h) // 2 + new_h, (w - new_w) // 2:(w - new_w) // 2 + new_w, :] = resized_image

    return canvas


def prep_image(img, inp_dim):
    """
    Prepare image for inputting to the neural network.

    Returns a Variable
    """
    if type(img) == str:
        orig_im = cv2.imread(img)
    else:
        orig_im = img
    dim = orig_im.shape[1], orig_im.shape[0]
    img = (letterbox_image(orig_im, (inp_dim, inp_dim)))
    img_ = img[:, :, ::-1].transpose((2, 0, 1)).copy()
    img_ = torch.from_numpy(img_).float().div(255.0).unsqueeze(0)
    return img_, orig_im, dim


def prep_image_pil(img, network_dim):
    orig_im = Image.open(img)
    img = orig_im.convert('RGB')
    dim = img.size
    img = img.resize(network_dim)
    img = torch.ByteTensor(torch.ByteStorage.from_buffer(img.tobytes()))
    img = img.view(*network_dim, 3).transpose(0, 1).transpose(0, 2).contiguous()
    img = img.view(1, 3, *network_dim)
    img = img.float().div(255.0)
    return img, orig_im, dim


def inp_to_image(inp):
    inp = inp.cpu().squeeze()
    inp = inp * 255
    try:
        inp = inp.data.numpy()
    except RuntimeError:
        inp = inp.numpy()
    inp = inp.transpose(1, 2, 0)

    inp = inp[:, :, ::-1]
    return inp



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

# ------------------------------------------------------------------------------
# Copyright (c) Microsoft
# Licensed under the MIT License.
# Written by Bin Xiao (Bin.Xiao@microsoft.com)
# ------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import cv2


def flip_back(output_flipped, matched_parts):
    '''
    ouput_flipped: numpy.ndarray(batch_size, num_joints, height, width)
    '''
    assert output_flipped.ndim == 4,\
        'output_flipped should be [batch_size, num_joints, height, width]'

    output_flipped = output_flipped[:, :, :, ::-1]

    # Input is a flipped image, so the output heatmaps need left/right joints swapped.
    for pair in matched_parts:
        tmp = output_flipped[:, pair[0], :, :].copy()
        output_flipped[:, pair[0], :, :] = output_flipped[:, pair[1], :, :]
        output_flipped[:, pair[1], :, :] = tmp

    return output_flipped


def fliplr_joints(joints, joints_vis, width, matched_parts):
    """
    flip coords
    """
    # Flip horizontal
    joints[:, 0] = width - joints[:, 0] - 1

    # Change left-right parts
    for pair in matched_parts:
        joints[pair[0], :], joints[pair[1], :] = \
            joints[pair[1], :], joints[pair[0], :].copy()
        joints_vis[pair[0], :], joints_vis[pair[1], :] = \
            joints_vis[pair[1], :], joints_vis[pair[0], :].copy()

    return joints*joints_vis, joints_vis


def transform_preds(coords, center, scale, output_size):
    target_coords = np.zeros(coords.shape)
    trans = get_affine_transform(center, scale, 0, output_size, inv=1)
    for p in range(coords.shape[0]):
        target_coords[p, 0:2] = affine_transform(coords[p, 0:2], trans)
    return target_coords


def get_affine_transform(
        center, scale, rot, output_size,
        shift=np.array([0, 0], dtype=np.float32), inv=0
):
    if not isinstance(scale, np.ndarray) and not isinstance(scale, list):
        print(scale)
        scale = np.array([scale, scale])

    scale_tmp = scale * 200.0
    src_w = scale_tmp[0]
    dst_w = output_size[0]
    dst_h = output_size[1]

    rot_rad = np.pi * rot / 180
    src_dir = get_dir([0, src_w * -0.5], rot_rad)
    dst_dir = np.array([0, dst_w * -0.5], np.float32)

    src = np.zeros((3, 2), dtype=np.float32)
    dst = np.zeros((3, 2), dtype=np.float32)
    src[0, :] = center + scale_tmp * shift
    src[1, :] = center + src_dir + scale_tmp * shift
    dst[0, :] = [dst_w * 0.5, dst_h * 0.5]
    dst[1, :] = np.array([dst_w * 0.5, dst_h * 0.5]) + dst_dir

    src[2:, :] = get_3rd_point(src[0, :], src[1, :])
    dst[2:, :] = get_3rd_point(dst[0, :], dst[1, :])

    if inv:
        trans = cv2.getAffineTransform(np.float32(dst), np.float32(src))
    else:
        trans = cv2.getAffineTransform(np.float32(src), np.float32(dst))

    return trans


def affine_transform(pt, t):
    new_pt = np.array([pt[0], pt[1], 1.]).T
    new_pt = np.dot(t, new_pt)
    return new_pt[:2]


def get_3rd_point(a, b):
    direct = a - b
    return b + np.array([-direct[1], direct[0]], dtype=np.float32)


def get_dir(src_point, rot_rad):
    sn, cs = np.sin(rot_rad), np.cos(rot_rad)

    src_result = [0, 0]
    src_result[0] = src_point[0] * cs - src_point[1] * sn
    src_result[1] = src_point[0] * sn + src_point[1] * cs

    return src_result


def crop(img, center, scale, output_size, rot=0):
    trans = get_affine_transform(center, scale, rot, output_size)

    dst_img = cv2.warpAffine(
        img, trans, (int(output_size[0]), int(output_size[1])),
        flags=cv2.INTER_LINEAR
    )

    return dst_img



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

import pytest
import torch
import torch.optim as optim
from fmpose3d.models import Model


class Args:
    """Mock args for model configuration."""
    channel = 512
    d_hid = 1024
    token_dim = 256
    layers = 5
    n_joints = 17


@pytest.fixture
def device():
    return torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')


@pytest.fixture
def model_args():
    return Args()


@pytest.fixture
def model(model_args, device):
    return Model(model_args).to(device)


@pytest.fixture
def optimizer(model):
    return optim.Adam(model.parameters(), lr=1e-3)


@pytest.fixture
def sample_batch(device):
    """Create a sample batch for training.
    
    Shapes based on FMPose3D_main.py train function:
    - input_2D: (B, F, J, 2) - 2D pose input
    - gt_3D: (B, F, J, 3) - 3D pose ground truth
    """
    batch_size = 4
    frames = 1
    joints = 17
    
    input_2D = torch.randn(batch_size, frames, joints, 2, device=device)
    gt_3D = torch.randn(batch_size, frames, joints, 3, device=device)
    
    return input_2D, gt_3D


def test_training_step_forward(model, sample_batch, device):
    """Test that a single forward pass works correctly."""
    input_2D, gt_3D = sample_batch
    B, F, J, _ = input_2D.shape
    
    # Create noise and interpolated sample (CFM training)
    x0 = torch.randn(B, F, J, 3, device=device)
    t = torch.rand(B, 1, 1, 1, device=device)
    y_t = (1.0 - t) * x0 + t * gt_3D
    
    # Forward pass
    v_pred = model(input_2D, y_t, t)
    
    # Check output shape
    assert v_pred.shape == (B, F, J, 3), f"Expected shape {(B, F, J, 3)}, got {v_pred.shape}"


def test_training_step_backward(model, optimizer, sample_batch, device):
    """Test that backward pass and gradient computation works."""
    input_2D, gt_3D = sample_batch
    B, F, J, _ = input_2D.shape
    
    # Create noise and interpolated sample
    x0 = torch.randn(B, F, J, 3, device=device)
    t = torch.rand(B, 1, 1, 1, device=device)
    y_t = (1.0 - t) * x0 + t * gt_3D
    v_target = gt_3D - x0
    
    # Forward pass
    v_pred = model(input_2D, y_t, t)
    
    # Compute loss (MSE)
    loss = ((v_pred - v_target) ** 2).mean()
    
    # Backward pass
    optimizer.zero_grad()
    loss.backward()
    
    # Check that gradients exist
    has_grad = False
    for param in model.parameters():
        if param.grad is not None and param.grad.abs().sum() > 0:
            has_grad = True
            break
    
    assert has_grad, "Model should have non-zero gradients after backward pass"


def test_training_step_parameter_update(model, optimizer, sample_batch, device):
    """Test that optimizer updates model parameters."""
    input_2D, gt_3D = sample_batch
    B, F, J, _ = input_2D.shape
    
    # Store initial parameter values
    initial_params = {name: param.clone() for name, param in model.named_parameters()}
    
    # Create noise and interpolated sample
    x0 = torch.randn(B, F, J, 3, device=device)
    t = torch.rand(B, 1, 1, 1, device=device)
    y_t = (1.0 - t) * x0 + t * gt_3D
    v_target = gt_3D - x0
    
    # Forward pass
    v_pred = model(input_2D, y_t, t)
    loss = ((v_pred - v_target) ** 2).mean()
    
    # Backward and update
    optimizer.zero_grad()
    loss.backward()
    optimizer.step()
    
    # Check that at least some parameters changed
    params_changed = False
    for name, param in model.named_parameters():
        if not torch.allclose(param, initial_params[name]):
            params_changed = True
            break
    
    assert params_changed, "Parameters should change after optimizer step"


def test_loss_decreases_over_steps(model, optimizer, sample_batch, device):
    """Test that loss decreases over multiple training steps."""
    input_2D, gt_3D = sample_batch
    B, F, J, _ = input_2D.shape
    
    losses = []
    num_steps = 10
    
    for _ in range(num_steps):
        # Create noise and interpolated sample
        x0 = torch.randn(B, F, J, 3, device=device)
        t = torch.rand(B, 1, 1, 1, device=device)
        y_t = (1.0 - t) * x0 + t * gt_3D
        v_target = gt_3D - x0
        
        # Forward pass
        v_pred = model(input_2D, y_t, t)
        loss = ((v_pred - v_target) ** 2).mean()
        losses.append(loss.item())
        
        # Backward and update
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
    
    # Check that loss generally decreases (last < first)
    # Note: May not always be strictly decreasing due to stochasticity
    assert losses[-1] < losses[0], f"Loss should decrease: first={losses[0]:.4f}, last={losses[-1]:.4f}"


def test_model_train_eval_modes(model, sample_batch, device):
    """Test that model can switch between train and eval modes."""
    input_2D, gt_3D = sample_batch
    B, F, J, _ = input_2D.shape
    
    x0 = torch.randn(B, F, J, 3, device=device)
    t = torch.rand(B, 1, 1, 1, device=device)
    y_t = (1.0 - t) * x0 + t * gt_3D
    
    # Test train mode
    model.train()
    assert model.training, "Model should be in training mode"
    v_pred_train = model(input_2D, y_t, t)
    assert v_pred_train.shape == (B, F, J, 3)
    
    # Test eval mode
    model.eval()
    assert not model.training, "Model should be in eval mode"
    with torch.no_grad():
        v_pred_eval = model(input_2D, y_t, t)
    assert v_pred_eval.shape == (B, F, J, 3)


"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

import glob

import cv2
import h5py
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm, colormaps
from scipy.optimize import least_squares
from tqdm import tqdm  # Import tqdm for the progress bar

joint_names = [
    "snout",
    "Right_Ear",
    "Left_Ear",
    "Shoulder_Center",
    "Right_Paw",
    "Right_Wrist",
    "Right_Elbow",
    "Right_Shoulder",
    "Left_Paw",
    "Left_Wrist",
    "Left_Elbow",
    "Left_Shoulder",
    "Body_Center",
    "Hip_Center",
    "Right_Foot",
    "Right_Ankle",
    "Right_Knee",
    "Left_Foot",
    "Left_Ankle",
    "Left_Knee",
    "Tail_Tip",
    "Tail_Middle",
    "Tail_Root",
]


# joint_names = [0'snout',
#                1'Right_Ear',
#                2'Left_Ear',
#                3'Shoulder_Center',
#                4'Right_Paw',
#                5'Right_Wrist',
#                6'Right_Elbow',
#                7'Right_Shoulder',
#                8'Left_Paw',
#                9'Left_Wrist',
#                10'Left_Elbow',
#                11'Left_Shoulder',
#                12'Body_Center',
#                13'Hip_Center',
#                14'Right_Foot',
#                15'Right_Ankle',
#                16'Right_Knee',
#                17'Left_Foot',
#                18'Left_Ankle',
#                19'Left_Knee',
#                20'Tail_Tip',
#                21'Tail_Middle',
#                22'Tail_Root']


# Skeleton connections
skeleton = [
    ["snout", "Right_Ear"],
    ["snout", "Left_Ear"],
    ["Shoulder_Center", "Right_Shoulder"],
    ["Right_Shoulder", "Right_Elbow"],
    ["Right_Elbow", "Right_Wrist"],
    ["Right_Wrist", "Right_Paw"],
    ["Shoulder_Center", "Left_Shoulder"],
    ["Left_Shoulder", "Left_Elbow"],
    ["Left_Elbow", "Left_Wrist"],
    ["Left_Wrist", "Left_Paw"],
    ["Shoulder_Center", "Body_Center"],
    ["Body_Center", "Hip_Center"],
    ["Hip_Center", "Right_Knee"],
    ["Right_Knee", "Right_Ankle"],
    ["Right_Ankle", "Right_Foot"],
    ["Hip_Center", "Left_Knee"],
    ["Left_Knee", "Left_Ankle"],
    ["Left_Ankle", "Left_Foot"],
    ["Tail_Root", "Tail_Middle"],
    ["Tail_Middle", "Tail_Tip"],
    ["Hip_Center", "Tail_Root"],
]

# from name to index
name_to_index = {name: idx for idx, name in enumerate(joint_names)}

# to skeleton
skeleton_indices = [[name_to_index[a], name_to_index[b]] for a, b in skeleton]  # start from 0


def compute_reprojection_errors(keypoints_2d, reprojected_2d):
    # Euclidean distance per keypoint
    errors = np.linalg.norm(keypoints_2d - reprojected_2d, axis=-1)  # (num_frames, num_keypoints)

    # Mean error over all frames/keypoints
    total_error = np.mean(errors)
    # Mean error per keypoint
    per_keypoint_error = np.mean(errors, axis=0)  # (num_keypoints,)

    return total_error, per_keypoint_error


import numpy as np


def compute_relative_errors(keypoints_2d, reprojected_2d):
    # Euclidean error (num_frames, num_keypoints)
    errors = np.linalg.norm(keypoints_2d - reprojected_2d, axis=-1)

    # Mean error
    total_error = np.mean(errors)

    # Mean error per keypoint (num_keypoints,)
    per_keypoint_error = np.mean(errors, axis=0)

    # Pairwise Euclidean distance between keypoints (num_frames, num_keypoints, num_keypoints)
    pairwise_dists = np.linalg.norm(
        keypoints_2d[:, :, None, :] - keypoints_2d[:, None, :, :], axis=-1
    )

    # Average inter-keypoint distance per frame
    avg_keypoint_distance = np.mean(pairwise_dists, axis=(1, 2))

    # Relative errors
    relative_error = total_error / np.mean(avg_keypoint_distance)

    # Relative error per keypoint (num_keypoints,)
    per_keypoint_relative_error = per_keypoint_error / np.mean(avg_keypoint_distance)

    return total_error, per_keypoint_error, relative_error, per_keypoint_relative_error


def normalize_points(points_3d):
    """Normalize 3D points to [-1, 1]."""
    min_vals = points_3d.min(axis=(0, 1), keepdims=True)
    max_vals = points_3d.max(axis=(0, 1), keepdims=True)
    points_3d_normalized = (points_3d - min_vals) / (max_vals - min_vals) * 2 - 1
    return points_3d_normalized


def plot_3d_keypoints_and_save_video(points_3d, output_video_path):

    num_points = len(joint_names)
    cmap = colormaps["rainbow"]
    colors = [cmap(i / num_points) for i in range(num_points)]

    # Create a VideoWriter to save the frames to a video
    fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    out = cv2.VideoWriter(
        output_video_path, fourcc, 30.0, (1024, 768)
    )  # Adjust the frame size if needed

    num_frames = len(points_3d)
    points_3d = normalize_points(points_3d)

    for frame in range(num_frames):
        fig = plt.figure(figsize=(10, 10))
        ax = fig.add_subplot(111, projection="3d")
        for i, joint in enumerate(joint_names):
            x, y, z = points_3d[frame][i]
            ax.scatter(x, y, z, color=colors[i], s=50)
        ax.set_xlabel("X")
        ax.set_ylabel("Y")
        ax.set_zlabel("Z")
        ax.set_title(f"Frame {frame+1}")

        # Adjust the view angle and limits to make the plot consistent
        ax.view_init(elev=30, azim=45)  # Adjust the view for better 3D perspective
        ax.set_xlim([-1, 1])  # Adjust based on your data range
        ax.set_ylim([-1, 1])
        ax.set_zlim([-1, 1])

        # Save the current figure as an image to be added to the video
        plt.subplots_adjust(left=0, right=1, top=1, bottom=0)  # Remove extra margins
        fig.canvas.draw()  # Draw the figure
        img = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        img = img.reshape(fig.canvas.get_width_height()[::-1] + (3,))  # Convert to RGB image format

        # Resize the image to fit video frame size
        img_resized = cv2.resize(img, (1024, 768))

        # Write the frame to the video
        out.write(img_resized)
        plt.close(fig)

        # Clear the figure to free memory
        plt.clf()

    out.release()


def plot_3d_skeleton_and_save_video(points_3d, output_video_path, num_frames_to_save=200):
    num_joints = len(joint_names)
    cmap = colormaps["rainbow"]
    colors = [cmap(i / num_joints) for i in range(num_joints)]

    fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    out = cv2.VideoWriter(output_video_path, fourcc, 30.0, (1920, 1080))

    # num_frames = min(points_3d.shape[0], num_frames_to_save)
    points_3d = normalize_points(points_3d)

    for frame in range(num_frames_to_save):
        fig = plt.figure(figsize=(10, 10))
        ax = fig.add_subplot(111, projection="3d")

        # draw keypoints
        for i, joint in enumerate(joint_names):
            # print("shape of 3d points",points_3d[frame][i])

            x, y, z = points_3d[frame][i]
            ax.scatter(x, y, z, color=colors[i], s=50)

        # draw skeleton connections
        for bone in skeleton:
            if bone[0] in joint_names and bone[1] in joint_names:
                i1, i2 = joint_names.index(bone[0]), joint_names.index(bone[1])
                x_vals = [points_3d[frame, i1, 0], points_3d[frame, i2, 0]]
                y_vals = [points_3d[frame, i1, 1], points_3d[frame, i2, 1]]
                z_vals = [points_3d[frame, i1, 2], points_3d[frame, i2, 2]]
                ax.plot(x_vals, y_vals, z_vals, color="black", linewidth=2, alpha=0.8)

        ax.set_xlabel("X")
        ax.set_ylabel("Y")
        ax.set_zlabel("Z")
        ax.set_title(f"Frame {frame+1}")
        ax.view_init(elev=30, azim=45)
        ax.set_xlim([-1, 1])
        ax.set_ylim([-1, 1])
        ax.set_zlim([-1, 1])

        # save frame and write to video
        fig.canvas.draw()
        img = np.frombuffer(fig.canvas.buffer_rgba(), dtype=np.uint8)
        img = img.reshape(fig.canvas.get_width_height()[::-1] + (4,))
        img_bgr = cv2.cvtColor(img, cv2.COLOR_RGBA2BGR)
        img_resized = cv2.resize(img_bgr, (1920, 1080))
        out.write(img_resized)

        plt.close(fig)

    out.release()


def load_camera_params(yaml_path, scale=0.5):
    """use opencv to read from yaml"""
    fs = cv2.FileStorage(yaml_path, cv2.FILE_STORAGE_READ)
    intrinsic_matrix = fs.getNode("intrinsicMatrix").mat()
    distortion_coeffs = fs.getNode("distortionCoefficients").mat()
    R = fs.getNode("R").mat()
    R = R.T
    T = fs.getNode("T").mat()
    fs.release()

    intrinsic_matrix = intrinsic_matrix.astype(np.float64)
    # check intrinsic matrix
    if not (np.allclose(intrinsic_matrix[2, :], [0, 0, 1], atol=1e-6)):
        intrinsic_matrix = intrinsic_matrix.T
    # scale the intrinsic matrix
    scale = 0.5
    if scale != 1.0:
        intrinsic_matrix[0, 0] *= scale
        intrinsic_matrix[1, 1] *= scale
        intrinsic_matrix[0, 2] *= scale
        intrinsic_matrix[1, 2] *= scale

    return {
        "intrinsic_matrix": intrinsic_matrix,
        "distortion_coeffs": distortion_coeffs,
        "R": R,
        "T": T,
    }


import cv2
import numpy as np


def triangulate_3d_batch(points_2d_batch, cameras):
    """
    input (num_cams, num_frames, num_joints, 2)

    parameters:
    - points_2d_batch: (6, num_frames, 23, 2)
    - cameras: list of length 6  `intrinsic_matrix`、`distortion_coeffs`、`R`、`T`

    return:
    - points_3d_batch: (num_frames, 23, 3)

    create matrix A to avoid for iteration
    """
    points_2d_batch = np.array(points_2d_batch)
    num_cams, num_frames, num_joints, _ = points_2d_batch.shape  # (6, num_frames, 23, 2)

    print("num_cams,num_frames,num_joinits", points_2d_batch.shape)
    # **1. compute projection matrics (6, 3, 4)**
    proj_matrices = np.array(
        [cam["intrinsic_matrix"] @ np.hstack((cam["R"], cam["T"])) for cam in cameras]
    )  # numpy array (6, 3, 4)

    # **2. undistortPoints**
    points_2d_undistorted = np.zeros_like(points_2d_batch)  # (6, num_frames, 23, 2)
    for i in range(num_cams):
        K, dist = cameras[i]["intrinsic_matrix"], cameras[i]["distortion_coeffs"]
        undistorted = cv2.undistortPoints(points_2d_batch[i].reshape(-1, 1, 2), K, dist, None, None)
        undistorted = undistorted.reshape(num_frames, num_joints, 2)
        print("undistorted shape:", undistorted.shape)
        print("ones shape:", np.ones((num_frames, num_joints, 1)).shape)
        # undistorted = (K @ np.hstack([undistorted, np.ones((num_frames, num_joints, 1))]).T).T[:, :, :2]
        undistorted = np.concatenate(
            [undistorted, np.ones((*undistorted.shape[:-1], 1))], axis=-1
        )  # (50, 23, 3)
        undistorted = (K @ undistorted[..., None])[..., 0]
        points_2d_undistorted[i] = undistorted[..., :2]

    # **3. Construct matrix A for triangulation**
    # Formula: A = [x P_3 - P_1; y P_3 - P_2], creating 6*2=12 equations per point
    x = points_2d_undistorted[..., 0]  # (6, num_frames, 23)
    y = points_2d_undistorted[..., 1]  # (6, num_frames, 23)

    # Extract projection matrix rows
    P1 = proj_matrices[:, None, None, 0, :]  # (6, 1, 1, 4)
    P2 = proj_matrices[:, None, None, 1, :]  # (6, 1, 1, 4)
    P3 = proj_matrices[:, None, None, 2, :]  # (6, 1, 1, 4)

    # Compute A (6, num_frames, 23, 2, 4)
    A = np.stack(
        [x[..., None] * P3 - P1, y[..., None] * P3 - P2], axis=-2
    )  # (6, num_frames, 23, 2, 4)
    A = A.reshape(num_cams * 2, num_frames, num_joints, 4)  # (12, num_frames, 23, 4)

    # **4. Solve using batch SVD**
    _, _, Vh = np.linalg.svd(A, full_matrices=False)  # Vh shape: (12, num_frames, 23, 4)
    X_hom = Vh[..., -1]  # Take last row (solution) (12, num_frames, 23, 4)

    # **5. Convert homogeneous coordinates to 3D**
    points_3d_batch = X_hom[..., :3] / X_hom[..., 3:]  # (num_frames, 23, 3)

    return points_3d_batch  # (num_frames, 23, 3)


def triangulate_3d(points_2d, cameras):
    # points_2d list of 6 in (23,2)
    """triangulate usd SVD"""
    proj_matrices = []
    points_2d_undistorted = []

    for i, cam in enumerate(cameras):
        K, dist, R, T = cam["intrinsic_matrix"], cam["distortion_coeffs"], cam["R"], cam["T"]

        P = K @ np.hstack((R, T))  # projectio matrix
        # print("Projection Matrix P:\n", P)

        proj_matrices.append(P)

        # undistortion
        # undistorted = cv2.undistortPoints(points_2d[i].reshape(-1, 1, 2), K, dist, None, K).reshape(-1, 2)
        undistorted = cv2.undistortPoints(
            points_2d[i].reshape(-1, 1, 2), K, dist, None, None
        ).reshape(-1, 2)
        undistorted = (K @ np.hstack([undistorted, np.ones((undistorted.shape[0], 1))]).T).T[:, :2]

        points_2d_undistorted.append(undistorted)

    # print("before undistortion and after",points_2d[0],points_2d_undistorted[0])
    # SVD
    num_points = points_2d_undistorted[0].shape[0]
    points_3d = np.zeros((num_points, 3))

    for j in range(num_points):
        A = []
        for i in range(len(proj_matrices)):
            P = proj_matrices[i]
            x, y = points_2d_undistorted[i][j]

            # build linear system Ax = 0
            A.append(x * P[2, :] - P[0, :])
            A.append(y * P[2, :] - P[1, :])

        A = np.array(A)
        _, _, Vh = np.linalg.svd(A)
        X_hom = Vh[-1]
        X = X_hom[:3] / X_hom[3]
        points_3d[j] = X

    # print("3D points and the 2D on camera 0",points_3d,points_2d_undistorted[0])
    return points_3d


def triangulate_3d_confi(points_2d, cameras):
    # points_2d: list of 6 in (23, 3), last dim is confidence
    """Triangulate using SVD with confidence"""
    proj_matrices = []
    points_2d_undistorted = []
    confidences = []

    for i, cam in enumerate(cameras):
        K, dist, R, T = cam["intrinsic_matrix"], cam["distortion_coeffs"], cam["R"], cam["T"]
        P = K @ np.hstack((R, T))  # Projection matrix
        proj_matrices.append(P)

        # Undistortion
        undistorted = cv2.undistortPoints(
            points_2d[i][:, :2].reshape(-1, 1, 2), K, dist, None, None
        ).reshape(-1, 2)
        undistorted = (K @ np.hstack([undistorted, np.ones((undistorted.shape[0], 1))]).T).T[:, :2]
        points_2d_undistorted.append(undistorted)

        # Collect confidences
        confidences.append(points_2d[i][:, 2])

    num_points = points_2d_undistorted[0].shape[0]
    points_3d = np.zeros((num_points, 4))  # last dimension stores confidence

    for j in range(num_points):
        A = []
        point_confidences = []
        for i in range(len(proj_matrices)):
            P = proj_matrices[i]
            x, y = points_2d_undistorted[i][j]
            conf = confidences[i][j]

            if conf > 0:
                # build linear system Ax = 0
                A.append(conf * (x * P[2, :] - P[0, :]))
                A.append(conf * (y * P[2, :] - P[1, :]))
                point_confidences.append(conf)

        if len(A) > 0:
            A = np.array(A)
            _, _, Vh = np.linalg.svd(A)
            X_hom = Vh[-1]
            X = X_hom[:3] / X_hom[3]
            points_3d[j, :3] = X

            # confidence: mean over valid views
            points_3d[j, 3] = np.mean(point_confidences)
        else:
            points_3d[j, 3] = 0

    return points_3d


# def project_3d_to_2d(points_3d, camera):
#     K, dist, R, T = camera["intrinsic_matrix"], camera["distortion_coeffs"], camera["R"], camera["T"]

#     points_3d = np.asarray(points_3d, dtype=np.float32)

#     R = np.array(R, dtype=np.float32)
#     T = np.array(T, dtype=np.float32)
#     K = np.array(K, dtype=np.float32)
#     dist = np.array(dist, dtype=np.float32)

#     points_2d_proj, _ = cv2.projectPoints(points_3d, R, T, K, dist)
#     return points_2d_proj.reshape(-1, 2)


def project_3d_to_2d(points_3d, camera):
    K, dist, R, T = (
        camera["intrinsic_matrix"],
        camera["distortion_coeffs"],
        camera["R"],
        camera["T"],
    )

    points_2d_proj, _ = cv2.projectPoints(points_3d, R, T, K, dist)
    return points_2d_proj.reshape(-1, 2)


def load_h5_keypoints(h5_path):
    """load 2D keypoint from h5 file"""
    with h5py.File(h5_path, "r") as f:
        group = f["df_with_missing"]
        dataset = group["block0_values"]
        # dataset = group['table']
        return np.array(dataset).reshape(-1, 23, 3)


def load_h5_keypoints_cspnext(h5_path):
    """load 2D keypoint from h5 file with cspnext format"""
    with h5py.File(h5_path, "r") as f:
        group = f["df_with_missing"]
        dataset = np.array(group["table"])
        values = dataset["values_block_0"]
        return values.reshape(-1, 23, 3)  # Reshape to (num_frames, num_joints, 3)


def visualize_2d_on_video(
    video_path, frame_number, original_keypoints, reprojected_keypoints, output_path
):
    """plot keypoint on frame n and save into png"""
    cap = cv2.VideoCapture(video_path)
    cap.set(cv2.CAP_PROP_POS_FRAMES, frame_number)  # jump to frame
    ret, frame = cap.read()

    if not ret:
        print(f"Failed to read frame {frame_number} from video {video_path}")
        cap.release()
        return

    h, w = frame.shape[:2]
    # reprojected_keypoints[:,0] /= h
    # reprojected_keypoints[:,1] /= w
    print("Original 2D Points (+):", original_keypoints)
    print("Projected 2D Points (dot):", reprojected_keypoints)
    print("height, and width", h, w)
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
    plt.figure(figsize=(10, 10))
    plt.imshow(frame)

    num_points = max(len(original_keypoints), len(reprojected_keypoints))
    cmap = colormaps["rainbow"]
    colors = [cmap(i / num_points) for i in range(num_points)]

    # plot original keypoints in +
    for i, point in enumerate(original_keypoints):
        if not np.isnan(point).any():
            x, y = point[0], point[1]
            if x < 0 or y < 0 or x >= w or y >= h:
                continue
            plt.scatter(
                x, y, marker="+", color=colors[i], s=15, linewidths=1, label=f"keypoints{i}"
            )

    # plot reprojected keypoints in o
    for i, point in enumerate(reprojected_keypoints):
        if not np.isnan(point).any():
            x, y = point[0], point[1]
            if x < 0 or y < 0 or x >= w or y >= h:
                continue
            plt.scatter(x, y, marker="o", color=colors[i], s=3, label=f"reprojected{i}")

    plt.axis("off")

    plt.savefig(output_path, bbox_inches="tight")
    plt.close()
    cap.release()


def main():
    yaml_files = sorted(
        glob.glob(
            "/workspace/MTFpose/data/Arber_data_noeva/076_02_221114_start0/calibration/*.yaml"
        )
    )
    h5_files = sorted(
        glob.glob("/workspace/MTFpose/data/Arber_data_noeva/076_02_221114_start0/pose2d_dlc/*.h5")
    )
    video_files = sorted(
        glob.glob("/workspace/MTFpose/data/Arber_data_noeva/076_02_221114_start0/video_dlc/*.mp4")
    )

    cameras = [load_camera_params(yaml) for yaml in yaml_files]

    # print("cameras len",len(cameras))
    # keypoints_2d = [load_h5_keypoints(h5) for h5 in h5_files]
    keypoints_2d = [load_h5_keypoints_cspnext(h5) for h5 in h5_files]

    print(
        "list: yaml ",
        len(yaml_files),
        "h5",
        len(h5_files),
        "video ",
        len(video_files),
        "cameras",
        len(cameras),
        "keypoints",
        len(keypoints_2d),
    )

    total_num_frames = keypoints_2d[0].shape[0]
    print("num_frames", total_num_frames, "for each cam", keypoints_2d[0].shape)  # (119498, 23, 3)

    # choose frame
    frame_number = 5

    # Choose the number of frames to visualize
    num_frames_to_save = total_num_frames

    points_3d_list = []
    repro_2d_list = []

    # save all 3d keypoints into .npy

    # visualization and save
    for frame in tqdm(
        range(total_num_frames), desc="Processing frames", unit="frame"
    ):  # loop in frames
        points_2d_frame = [keypoints_2d[cam_i][frame][:, :2] for cam_i in range(len(cameras))]

        # print("len of points 2d frame",len(points_2d_frame),points_2d_frame[0].shape) #len: 6 each shape :(23,2)

        # points_3d = triangulate_3d_confi(points_2d_frame, cameras)
        points_3d = triangulate_3d(points_2d_frame, cameras)
        points_3d_list.append(points_3d)

        # for i in range(len(cameras)):
        #     points_3d_array = np.array(points_3d_list).reshape(total_num_frames,-1,3)
        #     reprojected_2d = np.array([project_3d_to_2d(frame_3d, cameras[i]) for frame_3d in points_3d_array])
        #     total_error, per_keypoint_error, relative_error, per_keypoint_relative_error = compute_relative_errors(keypoints_2d[i][:total_num_frames,:,:2], reprojected_2d)

        # reprojected_2d = project_3d_to_2d(points_3d, cameras[0])
    points_3d = np.array(points_3d_list)
    print("shape of points 3d to save", points_3d.shape)  # (num_frames, 23, 3)
    output_npy_path = (
        "/workspace/MTFpose/data/Arber_data_noeva/076_02_221114_start0/triangulated_3d.npy"
    )
    np.save(output_npy_path, points_3d)
    # print("points_3d_fragment from 3d lift and stack, shape",points_3d.shape)

    #     if frame==frame_number:# visualize nth frame and save PNG
    #         for i in range(len(cameras)):
    #             output_path = f"/workspace/MTFpose/results/Camera_{i}_frame_{frame_number}.png"
    #             reprojected_2d = project_3d_to_2d(points_3d, cameras[i])
    #             # visualize_2d_on_video(video_files[i], frame_number, points_2d_frame[i], reprojected_2d, output_path)

    # # test on trangulate 3D batch
    # left_frame_id = 10
    # right_frame_id = 60

    # points_2d_fragment = [keypoints_2d[i][left_frame_id:right_frame_id,:,:2] for i in range(len(yaml_files))]

    # # get 3D keypoint from 3d lift
    # points_2d_fragment_np = np.array(points_2d_fragment) # (num_cams, num_frames, num_joints, 2) N,T,K,2
    # points_3d_fragment = triangulate_3d_batch(points_2d_fragment_np,cameras)  #(num_frames, 23, 3) T,K,3

    # # save skeleton for several frames
    # output_video_path = '/workspace/MTFpose/results/skeleton_batch_videodemo.mp4'
    # plot_3d_skeleton_and_save_video(points_3d_fragment,output_video_path,num_frames_to_save = num_frames_to_save)

    # compute reprojection error for each view
    for i in range(len(cameras)):
        # points_3d_list : (num_frames, num_keypoints, 3)
        # print("shape of points",len(points_3d_list),points_3d_list[0].shape)
        points_3d_array = np.array(points_3d_list).reshape(
            total_num_frames, -1, 3
        )  # with confidence

        # points_3d_array = points_3d_array[:,:,:3]
        # print("shape of points_3d_array",points_3d_array.shape,points_3d.dtype,points_3d_array)

        # reprojected_2d = project_3d_to_2d(points_3d_array, cameras[i])
        reprojected_2d = np.array(
            [project_3d_to_2d(frame_3d, cameras[i]) for frame_3d in points_3d_array]
        )
        # print("shape of reprojected_2d and points_3d",reprojected_2d.shape, points_3d_array.shape,keypoints_2d[0].shape)

        total_error, per_keypoint_error, relative_error, per_keypoint_relative_error = (
            compute_relative_errors(keypoints_2d[i][:total_num_frames, :, :2], reprojected_2d)
        )
        print(f"reprojection error in camera_{i}", total_error, relative_error)


if __name__ == "__main__":
    main()



"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""

"""
Common utilities for FMPose.
"""

from .arguments import opts
from .h36m_dataset import Human36mDataset
from .load_data_hm36 import Fusion
from .utils import (
    mpjpe_cal,
    p_mpjpe,
    AccumLoss,
    save_model,
    save_top_N_models,
    test_calculation,
    print_error,
    get_varialbe,
)

__all__ = [
    "opts",
    "Human36mDataset",
    "Fusion",
    "mpjpe_cal",
    "p_mpjpe",
    "AccumLoss",
    "save_model",
    "save_top_N_models",
    "test_calculation",
    "print_error",
    "get_varialbe",
]




"""
FMPose3D: monocular 3D Pose Estimation via Flow Matching

Official implementation of the paper:
"FMPose3D: monocular 3D Pose Estimation via Flow Matching"
by Ti Wang, Xiaohang Yu, and Mackenzie Weygandt Mathis
Licensed under Apache 2.0
"""


# ------------------------------------------------------------------------------
# Copyright (c) Microsoft
# Licensed under the MIT License.
# Written by Bin Xiao (Bin.Xiao@microsoft.com)
# ------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from yacs.config import CfgNode as CN


_C = CN()

_C.OUTPUT_DIR = ''
_C.LOG_DIR = ''
_C.DATA_DIR = ''
_C.GPUS = (0,)
_C.WORKERS = 4
_C.PRINT_FREQ = 20
_C.AUTO_RESUME = False
_C.PIN_MEMORY = True
_C.RANK = 0

# Cudnn related params
_C.CUDNN = CN()
_C.CUDNN.BENCHMARK = True
_C.CUDNN.DETERMINISTIC = False
_C.CUDNN.ENABLED = True

# common params for NETWORK
_C.MODEL = CN()
_C.MODEL.NAME = 'pose_hrnet'
_C.MODEL.INIT_WEIGHTS = True
_C.MODEL.PRETRAINED = ''
_C.MODEL.NUM_JOINTS = 17
_C.MODEL.TAG_PER_JOINT = True
_C.MODEL.TARGET_TYPE = 'gaussian'
_C.MODEL.IMAGE_SIZE = [256, 256]  # width * height, ex: 192 * 256
_C.MODEL.HEATMAP_SIZE = [64, 64]  # width * height, ex: 24 * 32
_C.MODEL.SIGMA = 2
_C.MODEL.EXTRA = CN(new_allowed=True)

_C.LOSS = CN()
_C.LOSS.USE_OHKM = False
_C.LOSS.TOPK = 8
_C.LOSS.USE_TARGET_WEIGHT = True
_C.LOSS.USE_DIFFERENT_JOINTS_WEIGHT = False

# DATASET related params
_C.DATASET = CN()
_C.DATASET.ROOT = ''
_C.DATASET.DATASET = 'mpii'
_C.DATASET.TRAIN_SET = 'train'
_C.DATASET.TEST_SET = 'valid'
_C.DATASET.DATA_FORMAT = 'jpg'
_C.DATASET.HYBRID_JOINTS_TYPE = ''
_C.DATASET.SELECT_DATA = False

# training data augmentation
_C.DATASET.FLIP = True
_C.DATASET.SCALE_FACTOR = 0.25
_C.DATASET.ROT_FACTOR = 30
_C.DATASET.PROB_HALF_BODY = 0.0
_C.DATASET.NUM_JOINTS_HALF_BODY = 8
_C.DATASET.COLOR_RGB = False

# train
_C.TRAIN = CN()

_C.TRAIN.LR_FACTOR = 0.1
_C.TRAIN.LR_STEP = [90, 110]
_C.TRAIN.LR = 0.001

_C.TRAIN.OPTIMIZER = 'adam'
_C.TRAIN.MOMENTUM = 0.9
_C.TRAIN.WD = 0.0001
_C.TRAIN.NESTEROV = False
_C.TRAIN.GAMMA1 = 0.99
_C.TRAIN.GAMMA2 = 0.0

_C.TRAIN.BEGIN_EPOCH = 0
_C.TRAIN.END_EPOCH = 140

_C.TRAIN.RESUME = False
_C.TRAIN.CHECKPOINT = ''

_C.TRAIN.BATCH_SIZE_PER_GPU = 32
_C.TRAIN.SHUFFLE = True

# testing
_C.TEST = CN()

# size of images for each device
_C.TEST.BATCH_SIZE_PER_GPU = 32
# Test Model Epoch
_C.TEST.FLIP_TEST = False
_C.TEST.POST_PROCESS = False
_C.TEST.SHIFT_HEATMAP = False

_C.TEST.USE_GT_BBOX = False

# nms
_C.TEST.IMAGE_THRE = 0.1
_C.TEST.NMS_THRE = 0.6
_C.TEST.SOFT_NMS = False
_C.TEST.OKS_THRE = 0.5
_C.TEST.IN_VIS_THRE = 0.0
_C.TEST.COCO_BBOX_FILE = ''
_C.TEST.BBOX_THRE = 1.0
_C.TEST.MODEL_FILE = ''

# debug
_C.DEBUG = CN()
_C.DEBUG.DEBUG = False
_C.DEBUG.SAVE_BATCH_IMAGES_GT = False
_C.DEBUG.SAVE_BATCH_IMAGES_PRED = False
_C.DEBUG.SAVE_HEATMAPS_GT = False
_C.DEBUG.SAVE_HEATMAPS_PRED = False


def update_config(cfg, args):
    cfg.defrost()
    cfg.merge_from_file(args.cfg)
    cfg.merge_from_list(args.opts)

    if args.modelDir:
        cfg.OUTPUT_DIR = args.modelDir

    # if args.logDir:
    #     cfg.LOG_DIR = args.logDir
    #
    # if args.dataDir:
    #     cfg.DATA_DIR = args.dataDir
    #
    # cfg.DATASET.ROOT = os.path.join(
    #     cfg.DATA_DIR, cfg.DATASET.ROOT
    # )
    #
    # cfg.MODEL.PRETRAINED = os.path.join(
    #     cfg.DATA_DIR, cfg.MODEL.PRETRAINED
    # )
    #
    # if cfg.TEST.MODEL_FILE:
    #     cfg.TEST.MODEL_FILE = os.path.join(
    #         cfg.DATA_DIR, cfg.TEST.MODEL_FILE
    #     )

    cfg.freeze()


if __name__ == '__main__':
    import sys
    with open(sys.argv[1], 'w') as f:
        print(_C, file=f)

